#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This module contains functions related with the execution
of the BLAST software (https://www.ncbi.nlm.nih.gov/books/NBK279690/).

Code documentation
------------------
"""


import subprocess

try:
    from utils import (constants as ct,
                       iterables_manipulation as im)
except ModuleNotFoundError:
    from CHEWBBACA.utils import (constants as ct,
                                 iterables_manipulation as im)


def make_blast_db(makeblastdb_path, input_fasta, output_path, db_type,
                  ignore=None):
    """Create a BLAST database.

    Parameters
    ----------
    makeblastdb_path : str
        Path to the 'makeblastdb' executable.
    input_fasta : str
        Path to the FASTA file that contains the sequences that
        will be added to the BLAST database.
    output_path : str
        Path to the directory where the database files will be
        created. Database files will have the same basename as
        the `input_fasta`.
    db_type : str
        Type of the database, nucleotide (nuc) or protein (prot).
    ignore : list or NoneType
        List with BLAST warnings that should be ignored.

    Returns
    -------
    stderr : list
        A list with the warnings and errors raised by BLAST.
    """
    # use '-parse-seqids' to be able to retrieve/align sequences by identifier
    blastdb_cmd = [makeblastdb_path, '-in', input_fasta,
                   '-out', output_path, '-parse_seqids',
                   '-dbtype', db_type]

    makedb_cmd = subprocess.Popen(blastdb_cmd,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE)

    stderr = makedb_cmd.stderr.readlines()

    # ignore errors/warnings provided to `ignore`
    if len(stderr) > 0:
        stderr = im.decode_str(stderr, 'utf8')
        if ignore is not None:
            stderr = im.filter_list(stderr, ignore)

    return stderr


def determine_blast_task(sequences, blast_type='blastp'):
    """Determine the type of BLAST task to execute.

    It is necessary to define the BLAST task if any of the
    sequences to align is shorter that 50 base pairs for
    BLASTn or 30 amino acids for BLASTp.

    Parameters
    ----------
    sequences : list
        List that contains strings representing DNA or
        protein sequences.
    blast_type : str
        Used to define the type of application, 'blastn'
        or 'blastp'.

    Returns
    -------
    blast_task : str
        A string that indicates the type of BLAST task to
        execute based on the minimum sequence size.

    Notes
    -----
    More information about the task option at:
        https://www.ncbi.nlm.nih.gov/books/NBK569839/
    """
    # get sequence length threshold for BLAST application
    length_threshold = ct.BLAST_TASK_THRESHOLD[blast_type]
    sequence_lengths = [len(p) for p in sequences]
    minimum_length = min(sequence_lengths)
    if minimum_length < length_threshold:
        blast_task = '{0}-short'.format(blast_type)
    else:
        blast_task = blast_type

    return blast_task


def run_blast(blast_path, blast_db, fasta_file, blast_output,
              max_hsps=1, threads=1, ids_file=None, blast_task=None,
              max_targets=None, ignore=None, composition_stats=None):
    """Execute BLAST to align sequences against a BLAST database.

    Parameters
    ----------
    blast_path : str
        Path to the BLAST application executable.
    blast_db : str
        Path to the BLAST database.
    fasta_file : str
        Path to the FASTA file with sequences to align against
        the BLAST database.
    blast_output : str
        Path to the file that will be created to store the
        results.
    max_hsps : int
        Maximum number of High Scoring Pairs per pair of aligned
        sequences.
    threads : int
        Number of threads/cores used to run BLAST.
    ids_file : str
        Path to a file with sequence identifiers, one per line.
        Sequences will only be aligned to the sequences in the
        BLAST database that match any of the identifiers in this
        file.
    blast_task : str
        Type of BLAST task.
    max_targets : int
        Maximum number of target/subject sequences to align
        against.
    ignore : list or None
        List with BLAST warnings that should be ignored.
    composition_stats : int
        Specify the composition-based statistics method used
        by BLAST.

    Returns
    -------
    stderr : list
        A list with the warnings and errors raised by BLAST.
    """
    # do not retrieve hits with high probability of occuring by chance
    blast_args = [blast_path, '-db', blast_db, '-query', fasta_file,
                  '-out', blast_output, '-outfmt', ct.BLAST_DEFAULT_OUTFMT,
                  '-max_hsps', str(max_hsps), '-num_threads', str(threads),
                  '-evalue', '0.001']

    # add file with list of sequence identifiers to align against
    if ids_file is not None:
        blast_args.extend(['-seqidlist', ids_file])
    # add type of BLASTp or BLASTn task
    if blast_task is not None:
        blast_args.extend(['-task', blast_task])
    # add maximum number of target sequences to align against
    if max_targets is not None:
        blast_args.extend(['-max_target_seqs', str(max_targets)])
    if composition_stats is not None:
        blast_args.extend(['-comp_based_stats', str(composition_stats)])

    blast_proc = subprocess.Popen(blast_args,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE)

    stderr = blast_proc.stderr.readlines()

    # ignore errors/warnings provided to `ignore`
    if len(stderr) > 0:
        stderr = im.decode_str(stderr, 'utf8')
        if ignore is not None:
            stderr = im.filter_list(stderr, ignore)

    return stderr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------
This module contains functions used to hash allelic profiles.

Code documentation
------------------
"""


import zlib
import hashlib

import pandas as pd
from Bio import SeqIO

try:
    from utils import (file_operations as fo,
                       iterables_manipulation as im,
                       multiprocessing_operations as mo)
except ModuleNotFoundError:
    from CHEWBBACA.utils import (file_operations as fo,
                                 iterables_manipulation as im,
                                 multiprocessing_operations as mo)


def hash_column(column, locus_file, hashing_function):
    """Substitute allele identifiers by allele sequence hashes.

    Parameters
    ----------
    column : pandas.core.series.Series
        Column with allele identifiers to substitute by the
        hash values computed from each allele sequence.
    locus_file : str
        Path to the FASTA file that contains the locus alleles.
    hashing_function : func
        Hashing function used to hash each allele.

    Returns
    -------
    hashed_column : pandas.core.series.Series
        Column where each allele identifier was substituted by
        the hash computed from the allele sequence.
    """
    # read Fasta files with locus alleles
    locus_alleles = {(rec.id).split('_')[-1].replace('*', ''): str(rec.seq)
                     for rec in SeqIO.parse(locus_file[0], 'fasta')}
    if len(locus_file) > 1:
        novel_records = {(rec.id).split('_')[-1].replace('*', ''): str(rec.seq)
                         for rec in SeqIO.parse(locus_file[1], 'fasta')}
        locus_alleles = im.merge_dictionaries([locus_alleles, novel_records], True)

    hashed_alleles = {}
    for seqid, seq in locus_alleles.items():
        # hash function does not accept string object, encode to get bytes object
        hashed_seq = hashing_function(seq.encode())
        # bitwise operation to convert crc32 and adler32 hashes to unsigned
        # integer and ensure the computed value is the same for Python 2 & 3
        if isinstance(hashed_seq, int):
            hashed_seq &= 0xffffffff
        else:
            hashed_seq = hashed_seq.hexdigest()
        hashed_alleles[seqid] = hashed_seq

    # faster than replace or map with update to avoid adding NaN
    hashed_column = column.apply(lambda x: hashed_alleles.get(x, x))

    return hashed_column


def hash_profiles(profiles_table, loci_ids, loci_files, hashing_function,
                  nrows, skiprows, output_directory):
    """Hash a set of allelic profiles read from a TSV file.

    Parameters
    ----------
    profiles_table : str
        Path to the TSV file that contains the allelic profiles.
    loci_ids : list
        List with the loci identifiers.
    loci_files : list
        List with the paths to the FASTA files that contain the
        loci alleles.
    hashing_function : func
        Hashing function used to hash each allele.
    nrows : int
        Number of rows/allelic profiles to read from the input
        file.
    skiprows : range
        Range of rows to skip.
    output_directory : str
        Path to the output directory.

    Returns
    -------
    output_file : str
        Path to the output file with the hashed profiles.
    """
    current_rows = pd.read_csv(profiles_table, delimiter='\t', dtype=str,
                               skiprows=skiprows, nrows=nrows, index_col=0)

    # remove all 'INF-' prefixes, missing data and '*' from identifiers
    current_rows = current_rows.apply(im.replace_chars, args=('-'))

    hashed_profiles = []
    for locus in loci_ids:
        locus_column = current_rows[locus]
        hashed_column = hash_column(locus_column, loci_files[locus],
                                    hashing_function)
        hashed_profiles.append(hashed_column)

    hashed_df = pd.concat(hashed_profiles, axis=1)
    start = skiprows.stop
    stop = skiprows.stop+len(hashed_df)-1
    input_basename = fo.file_basename(profiles_table, False)
    output_file = fo.join_paths(output_directory,
                                ['{0}_{1}-{2}_hashed.tsv'.format(input_basename, start, stop)])
    hashed_df.to_csv(output_file, sep='\t', index=True, header=False)

    return output_file


def main(profiles_table, schema_directory, output_directory, hash_type,
         cpu_cores, nrows, updated_files, no_inferred):

    # get hash function
    hashing_function = getattr(hashlib, hash_type, None)
    if hashing_function is None:
        hashing_function = getattr(zlib, hash_type, None)

    if hashing_function is None:
        print('{0} hash function is not available in '
              'hashlib or zlib modules.'.format(hash_type))
        return False

    # get loci identifiers
    with open(profiles_table, 'r') as infile:
        header = infile.readline()
        loci_ids = header.split()[1:]

    loci_files = {}
    for locus in loci_ids:
        locus_file = fo.join_paths(schema_directory, [locus])
        # add .fasta extension if file headers did not include it
        if locus_file.endswith('.fasta') is False:
            locus_file += '.fasta'
        loci_files[locus] = [locus_file]
        if locus_file in updated_files and no_inferred is True:
            loci_files[locus].append(updated_files[locus_file][0])

    # get input/sample identifiers
    sample_ids = pd.read_csv(profiles_table, delimiter='\t',
                             dtype=str, usecols=['FILE'])

    # write file with header
    header_basename = fo.file_basename(profiles_table).replace('.tsv', '_header.tsv')
    header_file = fo.join_paths(output_directory, [header_basename])
    fo.write_to_file(header, header_file, 'w', '')

    # create multiprocessing inputs
    multi_inputs = []
    # divide and process by row chunks
    for i in range(0, len(sample_ids), nrows):
        multi_inputs.append([profiles_table, loci_ids, loci_files,
                             hashing_function, nrows, range(1, i+1),
                             output_directory, hash_profiles])

    hashed_files = mo.map_async_parallelizer(multi_inputs, mo.function_helper,
                                             cpu_cores)

    # concatenate all files
    output_basename = fo.file_basename(profiles_table).replace('.tsv', '_hashed.tsv')
    output_file = fo.join_paths(output_directory, [output_basename])
    fo.concatenate_files([header_file]+hashed_files, output_file)

    # delete intermediate dataframes
    fo.remove_files([header_file]+hashed_files)

    return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This module enables the retrieval of information/stats from the
Chewie-NS. Its main objective is to provide information about
the list of species and schemas in the Chewie-NS, so that users
can quickly identify a schema of interest and download it (this
process generates tables with species and schemas identifiers that
can be passed to the `-sc` and `-sp` arguments of DownloadSchema).

Expected input
--------------

The process expects the following variables whether through command line
execution or invocation of the :py:func:`main` function:

- ``-m``, ``stats_mode`` : The process can retrieve the list of species
  ("species" option) in the Chewie-NS, the list of schemas for a species
  ("schemas" option and valid value for `--sp`) or information about a
  single schema ("schemas" option and valid values for `--sp` and `--sc`).

    - e.g.: ``species`` or ``schemas``

- ``--ns_url``, ``nomenclature_server_url`` : The base URL for the Nomenclature Server.
  The default value, "main", will establish a connection to "https://chewbbaca.online/",
  "tutorial" to "https://tutorial.chewbbaca.online/"" and "local" to
  "http://127.0.0.1:5000/NS/api/" (localhost). Users may also provide the IP address to
  other Chewie-NS instances.

    - e.g.: ``http://127.0.0.1:5000/NS/api/`` (localhost)

- ``--sp``, ``species_id`` : The integer identifier of a species
  in the Chewie-NS. The process will retrieve the list of schemas
  for the species with specified identifier.

    - e.g.: ``2``

- ``--sc``, ``schema_id`` : The integer identifier of a schema in
  the Chewie-NS. The process will retrieve information about the
  schema with specified identifier.

    - e.g.: ``4``

Code documentation
------------------
"""


import sys
import requests
import argparse
from urllib3.exceptions import InsecureRequestWarning

try:
    from utils import (constants as ct,
                       chewiens_requests as cr,
                       parameters_validation as pv)
except ModuleNotFoundError:
    from CHEWBBACA.utils import (constants as ct,
                                 chewiens_requests as cr,
                                 parameters_validation as pv)


# Suppress only the single warning from urllib3 needed.
requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)


def species_stats(base_url, headers_get):
    """Retrive summary statistics for a Chewie-NS instance.

    Retrieves the list of species in the Chewie-NS
    and the total number of schemas, loci and alleles
    per species.

    Parameters
    ----------
    base_url : str
        Base URL of the Chewie-NS.
    headers_get : dict
        HTTP headers for GET requests.

    Returns
    -------
    species_stats : list of str
        A list with formatted strings that can be joined
        and printed to display a table with the information
        retrieved from the Chewie-NS.
    """
    # get species and number of schemas
    species = species_schemas_count(base_url, headers_get)

    species_stats = []
    # table header
    species_stats.append('-'*78)
    species_stats.append('{:<30}  {:^10}  {:^10}  {:^10}  '
                         '{:^10}'.format('Species', 'id', '#schemas',
                                         '#loci', '#alleles'))
    species_stats.append('-'*78)

    for sp in species:
        # get totla number of loci and alleles per species
        stats = schema_stats(sp[0], base_url, headers_get)
        total_loci = 0
        total_alleles = 0
        for s in stats:
            total_loci += int(s['nr_loci'])
            total_alleles += int(s['nr_alleles'])

        species_stats.append('{:<30}  {:^10}  {:^10}  {:^10}  '
                             '{:^10}'.format(sp[1], sp[0], sp[2],
                                             total_loci, total_alleles))
    species_stats.append('-'*78)

    return species_stats


def schema_stats(species_id, base_url, headers_get):
    """Retrieve list of schemas for a species in Chewie-NS.

    Retrieves schema properties, number of loci and
    number of alleles for all schemas of a species in
    the Chewie-NS.

    Parameters
    ----------
    species_id : str
        The integer identifier of the species in
        the Chewie-NS.
    base_url : str
        Base URL of the Chewie-NS.
    headers_get : dict
        HTTP headers for GET requests.

    Returns
    -------
    res : list of dict or None
        List with one dict per schema ot NoneType
        if it was not possible to retrieve information.
    """
    endpoint_list = ['stats', 'species', species_id, 'totals']
    # unpack list of sequential endpoints and pass to create URI
    res = cr.simple_get_request(base_url, headers_get, endpoint_list)[1]
    status_code = res.status_code
    if status_code not in [200, 201]:
        res = None
    else:
        res = res.json()['message']

    return res


def species_schemas_count(base_url, headers_get):
    """Return the number of schemas per species in Chewie-NS.

    Parameters
    ----------
    base_url : str
        Base URL of the Chewie-NS.
    headers_get : dict
        HTTP headers for GET requests.

    Returns
    -------
    info : list of list
        A list with a sublist per species.
        Each sublist contains the species
        identifier, the name of the species
        and the total number of schemas.
    """
    endpoint_list = ['stats', 'species']
    # unpack list of sequential endpoints and pass to create URI
    res = cr.simple_get_request(base_url, headers_get, endpoint_list)[1]
    res = res.json()

    if 'message' in res:
        res = res['message']
    else:
        sys.exit('Could not retrieve species info.')

    if len(res) == 0:
        sys.exit('Could not retrieve species info.')
    else:
        info = []
        for s in res:
            sid = s['species']['value'].split('/')[-1]
            name = s['name']['value']
            num_schemas = s['schemas']['value']
            info.append([sid, name, num_schemas])

    # sort by species identifier
    info = sorted(info, key=lambda x: int(x[0]))

    return info


def species_schemas(species_id, base_url, headers_get):
    """Retrieve the list of schemas for a species in Chewie-NS.

    Parameters
    ----------
    species_id : str
        The integer identifier of the species in
        Chewie-NS.
    base_url : str
        Base URL of the Chewie-NS.
    headers_get : dict
        HTTP headers for GET requests.

    Returns
    -------
    res : list of dict
        The first dictionary contains the species
        URI and name and the following dictionaries
        contain the URI and name for all schemas
        associated with the species.
    """
    endpoint_list = ['species', species_id]
    # unpack list of sequential endpoints and pass to create URI
    res = cr.simple_get_request(base_url, headers_get, endpoint_list)[1]
    res = res.json()

    return res


def single_species(species_id, base_url, headers_get):
    """Retrieve summary statistics for a species' schemas.

    Retrieves the list of schemas for a species in
    a Chewie-NS instance and the total number of loci and
    alleles per schema.

    Parameters
    ----------
    species_id : str
        The integer identifier of the species in
        Chewie-NS.
    base_url : str
        Base URL for a Chewie-NS instance.
    headers_get : dict
        HTTP headers for GET requests.

    Returns
    -------
    schemas_stats : list of str
        A list with formatted strings that can be joined
        and printed to display a table with the information
        retrieved from Chewie-NS.

    Raises
    ------
    SystemExit
        - If the process cannot retrieve a species with
          specified identifier.
        - If the process could not retrieve schemas for the
          species.
    """
    # get all schemas for species
    schemas = species_schemas(species_id, base_url, headers_get)
    if 'NOT FOUND' in schemas:
        sys.exit('Could not find information about a species with '
                 'provided identifier (id={0}).'.format(species_id))
    species_name = schemas[0]['name']['value']
    schemas = {s['schemas']['value'].split('/')[-1]: s['schemaName']['value']
               for s in schemas[1:]}

    # get more info for all schema
    stats = schema_stats(species_id, base_url, headers_get)
    if stats is None:
        sys.exit('Could not retrieve schemas for {0} '
                 '(id={1}).'.format(species_name, species_id))
    schemas_stats = []
    schemas_stats.append('{0} (id={1})'.format(species_name, species_id))
    schemas_stats.append('-'*66)
    schemas_stats.append('{:<30}  {:^10}  {:^10}  {:^10}'
                         ''.format('Schema_name', 'id', '#loci', '#alleles'))
    schemas_stats.append('-'*66)
    for s in stats:
        schema_id = s['uri'].split('/')[-1]
        total_loci = s['nr_loci']
        total_alleles = s['nr_alleles']
        schema_name = schemas[schema_id]
        schemas_stats.append('{:<30}  {:^10}  {:^10}  {:^10}'
                             ''.format(schema_name, schema_id,
                                       total_loci, total_alleles))
    schemas_stats.append('-'*66)

    return schemas_stats


def single_schema(species_id, schema_id, base_url, headers_get):
    """Retrieve information about a schema in Chewie-NS.

    Parameters
    ----------
    species_id : str
        The integer identifier of the species in
        the Chewie-NS.
    schema_id : str
        The integer identifier of the schema in
        the Chewie-NS.
    base_url : str
        Base URL of the Chewie-NS.
    headers_get : dict
        HTTP headers for GET requests.

    Returns
    -------
    schema_info : list of str
        A list with formatted strings that can be joined
        and printed to display a table with the information
        retrieved from the Chewie-NS.

    Raises
    ------
    SystemExit
        - If the process cannot retrieve a species with
          specified identifier.
        - If the process cannot retrieve schemas for the
          species.
        - If the process cannot retrieve information about
          the schema with specified identifier.
    """
    # get all schemas for species
    species = species_schemas(species_id, base_url, headers_get)
    if 'NOT FOUND' in species:
        sys.exit('Could not find information about a species with '
                 'provided identifier (id={0}).'.format(species_id))
    species_name = species[0]['name']['value']

    schemas = schema_stats(species_id, base_url, headers_get)

    if schemas is None:
        sys.exit('Could not retrieve schemas for {0} '
                 '(id={1}).'.format(species_name, species_id))
    else:
        schema = [s for s in schemas if s['uri'].split('/')[-1] == schema_id]

    if len(schema) == 0:
        sys.exit('Could not find information about schema with '
                 'specified identifier (id={0}).'.format(schema_id))

    schema = schema[0]
    schema_info = []
    line_length = len(species_name)+len(schema['name'])+3
    schema_info.append('-'*line_length)
    schema_info.append('{0} - {1}'.format(species_name, schema['name']))
    schema_info.append('-'*line_length)

    schema_info.append('\nID: {0}'
                       ''.format(schema['uri'].split('/')[-1]))
    schema_info.append('Created by: {0}'
                       ''.format(schema['user']))
    schema_info.append('Total loci: {0}'
                       ''.format(schema['nr_loci']))
    schema_info.append('Total alleles: {0}'
                       ''.format(schema['nr_alleles']))
    schema_info.append('BLAST Score Ratio: {0}'
                       ''.format(schema['bsr']))
    schema_info.append('chewBBACA version: {0}'
                       ''.format(schema['chewBBACA_version']))
    schema_info.append('Genetic code: {0}'
                       ''.format(schema['translation_table']))
    schema_info.append('Minimum sequence length: {0}'
                       ''.format(schema['minimum_locus_length']))
    schema_info.append('Sequence length variation threshold: {0}'
                       ''.format(schema['size_threshold']))
    schema_info.append('Clustering word size: {0}'
                       ''.format(schema['word_size']))
    schema_info.append('Clustering similarity: {0}'
                       ''.format(schema['cluster_sim']))
    schema_info.append('Representative similarity filter: {0}'
                       ''.format(schema['representative_filter']))
    schema_info.append('Intracluster similarity filter: {0}'
                       ''.format(schema['intraCluster_filter']))
    schema_info.append('Creation date: {0}'
                       ''.format(schema['dateEntered']))
    schema_info.append('Last modified: {0}'
                       ''.format(schema['last_modified']))

    return schema_info


def main(mode, nomenclature_server, species_id, schema_id):

    headers_get = ct.HEADERS_GET_JSON

    print('\nRetrieving data...')
    if mode == 'species':
        stats = species_stats(nomenclature_server, headers_get)
    elif mode == 'schemas':
        if species_id is not None:
            if schema_id is not None:
                stats = single_schema(species_id, schema_id,
                                      nomenclature_server, headers_get)
            else:
                stats = single_species(species_id, nomenclature_server,
                                       headers_get)
        else:
            sys.exit('\nPlease provide a valid species identifier '
                     'to get the list of available schemas.\n')

    # print stats
    stats_text = '\n'.join(stats)
    print('\n{0}\n'.format(stats_text))


def parse_arguments():

    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-m', type=str, required=True,
                        dest='stats_mode', choices=['species', 'schemas'],
                        help='The process can retrieve the list of species '
                             '("species" option) in the Chewie-NS or the '
                             'list of schemas for a species '
                             '("schemas" option).')

    parser.add_argument('--ns', type=pv.validate_ns_url, required=False,
                        default='main',
                        dest='nomenclature_server',
                        help='The base URL for the Nomenclature Server. '
                             'The default value, "main", will establish a '
                             'connection to "https://chewbbaca.online/", '
                             '"tutorial" to "https://tutorial.chewbbaca.online/" '
                             'and "local" to "http://127.0.0.1:5000/NS/api/" (localhost). '
                             'Users may also provide the IP address to other '
                             'Chewie-NS instances.')

    parser.add_argument('--sp', type=str, required=False,
                        dest='species_id', default=None,
                        help='The integer identifier of a '
                             'species in the Chewie-NS.')

    parser.add_argument('--sc', type=str, required=False,
                        dest='schema_id', default=None,
                        help='The integer identifier of a schema in '
                             'the Chewie-NS.')

    args = parser.parse_args()

    return args


if __name__ == '__main__':

    args = parse_arguments()
    main(**(vars(args)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This module joins allele calling results from different
runs. It can concatenate files with allelic profiles
for the same set of loci or create a new file with the
allelic profiles for the loci that were common between
all input files.

Code documentation
------------------
"""


import os
import sys

import pandas as pd

try:
    from utils import file_operations as fo
except ModuleNotFoundError:
    from CHEWBBACA.utils import file_operations as fo


def concatenate_profiles(files, loci_list, output_file):
    """Concatenate allele calling results for a common set of loci.

    Parameters
    ----------
    files : list
        List with the paths to the TSV files with
        allele calling results.
    loci_list : list
        List with the identifiers of common loci.
    output_file : str
        Path to the output file.

    Returns
    -------
    total_profiles : int
        Number of profiles written to the output file.
    """
    total_profiles = 0
    for f in files:
        # create dataframe with columns for loci in list
        profiles = pd.read_csv(f, sep='\t',
                               usecols=loci_list, dtype=str)
        # reorder columns to prevent concatenating results
        # with same loci but with different loci/column order
        profiles = profiles[loci_list]
        total_profiles += len(profiles)
        # save dataframe to file
        profiles.to_csv(output_file, mode='a',
                        header=not os.path.exists(output_file),
                        sep='\t', index=False)

    return total_profiles


def main(profiles, output_file, common):

    if len(profiles) == 1:
        sys.exit('Provided a single file. Nothing to do.')

    headers = []
    for file in profiles:
        header = fo.read_lines(file, strip=True, num_lines=1)
        headers.append(header[0].split('\t'))

    if common is False:
        # check if headers are equal
        if all([set(headers[0]) == set(h) for h in headers[1:]]) is True:
            print('Profiles have {0} loci.'.format(len(headers[0])-1))

            total_profiles = concatenate_profiles(profiles,
                                                  headers[0],
                                                  output_file)
        else:
            sys.exit('Files have different sets of loci. Please provide '
                     'files with the results for the same set of loci or '
                     'provide the "--common" parameter to create a file '
                     'with the results for the set of common loci.')
    else:
        # determine set of common loci
        common_loci = headers[0]
        for h in headers[1:]:
            # determine common loci ordered based on headers in first file
            latest_common = [locus for locus in common_loci if locus in h]
            # update set of common loci so far
            common_loci = latest_common

        if len(latest_common) <= 1:
            sys.exit('Profiles do not have loci in common.')

        print('Profiles have {0} loci in common.'
              ''.format(len(latest_common)-1))
        total_profiles = concatenate_profiles(profiles,
                                              latest_common,
                                              output_file)

    print('Joined {0} files with a total of {1} '
          'profiles.'.format(len(profiles), total_profiles))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This module contains functions to work with iterables.

Code documentation
------------------
"""


import re
import hashlib
import itertools

try:
    from utils import (constants as ct,
                       fasta_operations as fao)
except ModuleNotFoundError:
    from CHEWBBACA.utils import (constants as ct,
                                 fasta_operations as fao)


def match_regex(text, pattern):
    """Extract substrings that match a regex pattern.

    Parameters
    ----------
    text : str
        Input text to search the pattern in.
    pattern : str
        Regular expression pattern.

    Returns
    -------
    match : str
        Matched substring or NoneType if no matches
        were found.
    """
    match = re.search(pattern, text)
    if match is not None:
        match_span = match.span()
        match = text[match_span[0]:match_span[1]]

    return match


def join_list(lst, delimiter):
    """Join all elements in a list into a single string.

    Parameters
    ----------
    lst : list
        List with elements to be joined.
    delimiter : str
        Character used to join list elements.

    Returns
    -------
    joined_list : str
        A single string with all elements in the input
        list joined by the character chosen as link.
    """
    joined_list = delimiter.join(lst)

    return joined_list


def flatten_list(list_to_flatten):
    """Flatten one level of a nested list.

    Parameters
    ----------
    list_to_flatten : list
        Nested list to flatten.

    Returns
    -------
    flattened_list : str
        Input list flattened by one level.
    """
    flattened_list = list(itertools.chain(*list_to_flatten))

    return flattened_list


def isListEmpty(input_list):
    """Check if a nested list is empty."""
    if isinstance(input_list, list):
        return all(map(isListEmpty, input_list)) if isinstance(input_list, list) else False


def divide_list_into_n_chunks(list_to_divide, n):
    """Divides a list into a defined number of sublists.

    Parameters
    ----------
    list_to_divide : list
        List to divide into sublists.
    n : int
        Number of sublists to create.

    Returns
    -------
    sublists : list
        List with the sublists created by dividing
        the input list.
    """
    sublists = []
    d, r = divmod(len(list_to_divide), n)
    for i in range(n):
        si = (d+1)*(i if i < r else r) + d*(0 if i < r else i - r)
        sublists.append(list_to_divide[si:si+(d+1 if i < r else d)])

    # exclude lists that are empty due to small number of elements
    sublists = [line for line in sublists if len(line) > 0]

    return sublists


def merge_dictionaries(dictionaries_list, overwrite=False):
    """Merge several dictionaries.

    Parameters
    ----------
    dictionaries_list : list
        A list with the dictionaries to merge.
    overwrite : bool
        True to overwrite values when there is a key
        collision, False otherwise.

    Returns
    -------
    merged_dicts : dict
        A dictionary resulting from merging
        all input dictionaries.
    """
    merged_dicts = dictionaries_list[0]
    if overwrite is True:
        for d in dictionaries_list[1:]:
            merged_dicts = {**merged_dicts, **d}
    elif overwrite is False:
        for d in dictionaries_list[1:]:
            for k, v in d.items():
                if k in merged_dicts:
                    merged_dicts[k] = list(set.union(set(v), set(merged_dicts[k])))
                else:
                    merged_dicts[k] = v

    return merged_dicts


def invert_dictionary(dictionary):
    """Invert a dictionary (key:value to value:key).

    Parameters
    ----------
    dictionary : dict
        Dictionary to be inverted.

    Returns
    -------
    inverted_dict : dict
        Inverted dictionary.
    """
    inverted_dict = {value: key for key, value in dictionary.items()}

    return inverted_dict


def split_iterable(iterable, size):
    """Split a dictionary.

    Parameters
    ----------
    iterable : dict
        Dictionary to split.
    size : int
        Size of dictionaries created from the input
        dictionary.

    Returns
    -------
    chunks : list
        List with dictionaries of defined size
        resulting from splitting the input dictionary.
    """
    chunks = []
    # create iterable from input object
    # need to convert or slicing will always return the first elements
    it = iter(iterable)
    for i in range(0, len(iterable), size):
        chunks.append({k: iterable[k] for k in itertools.islice(it, size)})

    return chunks


def select_keys(dictionary, size):
    """Select dictionary keys based on the number of values.

    Parameters
    ----------
    dictionary : dict
        Dictionary with key:value pairs to select.
    size : int
        Number of values used to select dictionary entries.

    Returns
    -------
    selected : list
        List with the dictionary keys that have a number
        of matching values equal or greater/lesser than
        `size`.
    """
    selected = [k for k, v in dictionary.items() if len(v) == size]

    return selected


def prune_dictionary(dictionary, keys):
    """Remove a set of keys from a dictionary.

    Parameters
    ----------
    dictionary : dict
        Input dictionary.
    keys : list
        List of keys to remove.

    Returns
    -------
    pruned_dict : dict
        Dictionary without the keys in the list
        of keys to remove.
    """
    pruned_dict = {k: v for k, v in dictionary.items() if k not in keys}

    return pruned_dict


def extract_subsequence(sequence, start, stop):
    """Extract a substring from a string.

    Parameters
    ----------
    sequence : str
        Input string.
    start : int
        Substring start position in input string.
    stop : int
        Substring stop position in input string.

    Returns
    -------
    subsequence : str
        Substring extracted from input string.
    """
    subsequence = sequence[start:stop]

    return subsequence


def extract_single_cds(sequence, start, stop, strand):
    """Extract a coding sequence from a larger sequence.

    Parameters
    ----------
    sequence : str
        Contig that contains the coding sequence.
    start : int
        Coding sequence start position in contig.
    stop : int
        Coding sequence stop position in contig.
    strand : int
        Coding sequence orientation.

    Returns
    -------
    coding_sequence : str
        Coding sequence extracted from input contig.
    """
    coding_sequence = extract_subsequence(sequence, start, stop)

    if strand == 0:
        coding_sequence = reverse_complement(coding_sequence, ct.DNA_BASES)

    return coding_sequence


def escape_special_characters(input_string):
    """Escape strings to use in regex.

    Parameters
    ----------
    input_string : str
        String containing characters to escape.

    Returns
    -------
    escaped_string : str
        Escaped string.
    """
    escaped_string = re.escape(input_string)

    return escaped_string


def replace_multiple_characters(input_string, replacements):
    """Replace multiple characters in a string.

    Parameters
    ----------
    input_string : str
        String with characters to be replaced.
    replacements : list
        List that contains sublists with two elements,
        the element to be replaced and the element it should
        be replaced by.

    Returns
    -------
    replaced : str
        Input string with replaced characters.
    """
    for r in replacements:
        if r[0] in input_string:
            input_string = input_string.replace(*r)

    return input_string


def reverse_complement(input_string, alphabet):
    """Determine the reverse complement of a string.

    Parameters
    ----------
    input_string : str
        Input string.
    alphabet : str
        String that contains the alphabet used in
        the input string (the reverse must be equal
        to the desired complement).

    Returns
    -------
    reverse_complement_string : str
        The reverse complement of the input string.
    """
    translation_table = str.maketrans(alphabet, alphabet[::-1])

    upper_string = input_string.upper()
    complement_string = upper_string.translate(translation_table)
    reverse_complement_string = reverse_str(complement_string)

    return reverse_complement_string


def reverse_str(input_string):
    """Reverse character order in a string.

    Parameters
    ----------
    input_string : str
     String to be reversed.

    Returns
    -------
    revstr : str
        Reverse of input string.
    """
    revstr = input_string[::-1]

    return revstr


def check_str_alphabet(input_string, alphabet):
    """Determine if a string only contains characters from specified alphabet.

    Parameters
    ----------
    input_string : str
        Input string.
    alphabet : str
        String that includes all characters in the
        alphabet.

    Returns
    -------
    True if sequence only has characters from specified
    alphabet, False otherwise.
    """
    alphabet_chars = set(alphabet)
    string_chars = set(input_string)

    diff = string_chars - alphabet_chars

    return len(diff) == 0


def check_str_multiple(input_string, number):
    """Determine if length of input string is a multiple of a specified number.

    Parameters
    ----------
    input_string : str
        Input string.
    number : int
        Length value should be a multiple of this number.

    Returns
    -------
    True if the length of the sequence is a multiple of the
    specified number, False otherwise.
    """
    return (len(input_string) % number) == 0


def hash_sequence(input_string, hash_type='sha256'):
    """Compute hash of an input string.

    Parameters
    ----------
    input_string : str
        Input string to hash.
    hash_type : str
        Hash type/function that will be used to compute the
        hash (any of the hash functions available in the
        hashlib module).

    Returns
    -------
    hashed_string : str
        String representation of the HASH object
        in hexadecimal digits.
    """
    # get hash function object from hashlib
    hashing_function = getattr(hashlib, hash_type)

    # default encoding is UTF-8
    hashed_string = hashing_function(input_string.encode()).hexdigest()

    return hashed_string


def string_kmerizer(input_string, k_value, offset=1, position=False):
    """Decompose a string into k-mers.

    Parameters
    ----------
    input_string : str
        String to divide into k-mers.
    k_value : int
        Value for the size of k-mers.
    offset : int
        Value to indicate offset of consecutive k-mers.
    position : bool
        If the start position of the k-mers in the string
        should be stored.

    Returns
    -------
    kmers : list
        List that contains the k-mers determined for the
        input string. The list will contain strings if
        it is not specified that positions should be
        stored and tuples of k-mer and start position
        if the position is stored.
    """
    if position is False:
        kmers = [input_string[i:i+k_value]
                 for i in range(0, len(input_string)-k_value+1, offset)]
    elif position is True:
        kmers = [(input_string[i:i+k_value], i)
                 for i in range(0, len(input_string)-k_value+1, offset)]

    return kmers


def determine_minimizers(input_string, adjacent_kmers, k_value, offset=1,
                         position=False):
    """Determine minimizers for a input string.

    Determines minimizers for a string based on lexicographical
    order. Skips windows that cannot have a minimizer based on
    the minimizer computed in the previous iteration.

    Parameters
    ----------
    input_string : str
        String representing the sequence.
    adjacent_kmers : int
        Window size value. Number of adjacent k-mers per group.
    k_value : int
        Value of k for the k-mer size.
    offset : int
        Value to indicate offset of consecutive k-mers.
    position : bool
        If the start position of the k-mers in the sequence
        should be stored.

    Returns
    -------
    minimizers : list
        A list with the set of minimizers determined
        for the input string.
    """
    # break string into k-mers
    kmers = string_kmerizer(input_string, k_value, offset, position)

    i = 0
    previous = None
    sell = False
    minimizers = []
    # determine total number of windows
    last_window = (len(kmers)-adjacent_kmers)
    while i <= last_window:
        # get kmers in current window
        window = kmers[i:i+adjacent_kmers]
        # pick smallest kmer as minimizer
        minimizer = [min(window)]
        # get position in window of smallest minimizer
        minimizer_idx = window.index(minimizer[0])
        # sliding window that does not included last minimizer
        if previous is None:
            # simply store smallest minimizer
            minimizers.extend(minimizer)
        # sliding window includes last minimizer because we
        # skipped some sliding windows
        else:
            # check if minimizer is the same as the one picked
            # in the last window
            # Do not store minimizer if it is the same
            if minimizer[0] != previous:
                # get kmers smaller than last minimizer
                skipped = window[1:minimizer_idx]
                # determine if any of the smaller kmers is
                # the minimizer of a skipped window
                minimal = previous
                for m in skipped:
                    if m < minimal:
                        minimizer.append(m)
                        minimal = m
                minimizers.extend(minimizer)

        # slide by 1 if minimizer has index 0 in window
        if minimizer_idx == 0:
            i += 1
            previous = None
        # skip sliding windows based on minimizer position
        else:
            i += minimizer_idx
            # if adding minimizer index surpasses last window value we
            # might miss one last minimizer because it will fail the condition
            # find a better way to control this condition!
            if i > last_window and sell is False:
                i = last_window
                sell = True
            previous = minimizer[0]

    return minimizers


def decode_str(str_list, encoding):
    """Decode bytes objects.

    Decodes bytes objects in the input list and strips decoded
    strings from whitespaces and newlines.

    Parameters
    ----------
    str_list
        List with string or bytes objects to decode
        and strip of whitespaces and newlines.
    encoding : str
        Encoding codec to use.

    Returns
    -------
    decoded : list
        List with strings without whitespaces or
        newlines.
    """
    decoded = [m.decode(encoding).strip()
               if type(m) == bytes
               else m.strip()
               for m in str_list]

    return decoded


# sorted key parameter is None by default
def sort_iterable(data, sort_key=None, reverse=False):
    """Sort an iterable.

    Parameters
    ----------
    data : iter
        Iterable to sort.
    sort_key
        If provided, data will be sorted based
        on this function.
    reverse : bool
        If sorting order should be inverted.

    Returns
    -------
    sorted_data
        Iterable with sorted elements.
    """
    # sorted key parameter is None by default
    sorted_data = sorted(data, key=sort_key, reverse=reverse)

    return sorted_data


def filter_list(lst, remove):
    """Remove elements from a list.

    Parameters
    ----------
    lst : list
        Input list.
    remove : list
        List of elements to remove from input list.

    Returns
    -------
    filtered_list : list
        List without the removed elements.
    """
    filtered_list = list(set(lst) - set(remove))

    return filtered_list


def find_missing(integer_list):
    """Find the set of integers missing from a list to make it consecutive.

    Parameters
    ----------
    integer_list : list
        List containing integers.

    Returns
    -------
    missing_integers : list
        Sorted list of missing integers.
    """
    first = integer_list[0]
    last = integer_list[-1]
    missing_integers = sorted(set(range(first, last + 1)).difference(integer_list))

    return missing_integers


# add new parameter that accepts method to sample kmers
def kmer_index(fasta_file, word_size, fasta=True):
    """Create a k-mer index from a set of sequences in a FASTA file.

    Parameters
    ----------
    fasta_file : str
        Path to a Fasta file.
    word_size : int
        Value k for the k-mer size.
    fasta : bool
        True if input is a FASTA file, False if it is a dictionary
        with sequence identifiers as keys and sequences as values.

    Returns
    -------
    kmers_mapping : dict
        Dictionary with k-mers as keys and the
        list of sequence identifiers of the
        sequences that contain the kmers as
        values.
    seqs_kmers : dict
        Dictionary with sequence identifiers
        as keys and the set of distinct k-mers
        for each sequence as values.
    """
    if fasta is True:
        sequences = fao.sequence_generator(fasta_file)
    else:
        sequences = fasta_file

    kmers_mapping = {}
    for record in sequences:
        try:
            seqid = record
            sequence = sequences[record]
        except Exception as e:
            seqid = record.id
            sequence = str(record.seq)

        minimizers = determine_minimizers(sequence, word_size,
                                          word_size, position=False)
        kmers = set(minimizers)

        # create dict with kmers as keys and list
        # of sequences with given kmers as values
        for kmer in kmers:
            kmers_mapping.setdefault(kmer, []).append(seqid)

    return kmers_mapping


def contained_terms(iterable, terms):
    """Find elements in an iterable that contain any term from a set of terms.

    Parameters
    ----------
    iterable
        An iterable such as a list with strings
        or a dictionary.
    terms : list
        Terms to search for.

    Returns
    -------
    matches : list
        List with the elements of the iterable
        that contain any of the searched terms.
    """
    matches = []
    for e in iterable:
        if any([term in e for term in terms]) is True:
            matches.append(e)

    return matches


def integer_mapping(values, inverse=False):
    """Create mapping between list elements and consecutive integers.

    Parameters
    ----------
    values : iter
        Input iterable.
    inverse : bool
        Invert mapping order, integers:values instead
        of values:integers.

    Returns
    -------
    mapping : dict
        Dictionary with the mapping between the
        elements in the input iterable and the
        sequential integers.
    """
    mapping = {v: i+1 for i, v in enumerate(values)}
    if inverse is True:
        mapping = invert_dictionary(mapping)

    return mapping


def multiprocessing_inputs(inputs, common_args, function):
    """Create lists of inputs to process in parallel.

    Creates lists of inputs for the `map_async_parallelizer`
    function from the `multiprocessing_operations` module.

    Parameters
    ----------
    inputs : list
        List with the distinct inputs.
    common_args : list
        List with the common arguments/inputs.
    function : func
        Function that will be parallelized by the
        `map_async_parallelizer` function.

    Returns
    -------
    input_groups : list
        List with one sublist per distinct input. Each
        sublist also contains the common arguments and
        the function that will receive the arguments.
    """
    input_groups = []
    # create a list for each distinct input
    for g in inputs:
        new_input = g + common_args + [function]
        input_groups.append(new_input)

    return input_groups


def aggregate_iterables(iterables):
    """Aggregate elements with same index from several iterables.

    Parameters
    ----------
    iterables : list
        List of iterables to aggreagate.

    Returns
    -------
    aggregated_inputs : list
        List with a sublist per group of elements
        with same index that were aggregated.
    """
    aggregated_inputs = [list(i) for i in zip(*iterables)]

    return aggregated_inputs


def mapping_function(values, function, args):
    """Create mapping between input elements and function result.

    Parameters
    ----------
    values : iter
        Input iterable.
    function : func
        Funtion to apply to the elements in the
        input iterable.
    args : list
        List of arguments to pass to the function.

    Returns
    -------
    mapping : dict
        Dictionary with the mapping between the
        elements in the input iterable and the
        result of applying the function to each
        value.
    """
    mapping = {v: function(v, *args) for v in values}

    return mapping


def polyline_encoding(number_list, precision=0):
    """Use the polyline algorithm to encode/compress a list of numbers.

    Parameters
    ----------
    number_list : list
        List with numbers to encode.
    precision : int
        Number of decimal places preserved by the encoding.

    Returns
    -------
    compressed_values : str
        A single string composed of ASCII characters
        that represents the compressed list of numbers
        to the desired level of precision.

    Notes
    -----
    This implementation is based on the `numcompress` package
    (https://github.com/amit1rrr/numcompress).
    """
    compressed_values = ''
    # store the precision value
    # to ensure proper character display, encoded values are summed with 63
    # (the ASCII character '?')
    compressed_values += chr(precision+63)
    previous_num = 0
    for number in number_list:
        # encode the difference between numbers
        difference = number - previous_num
        # multiply and round to get integer that preserves decimal places
        difference = int(round(difference*(10**precision)))
        # left bitwise shift 1 position (0b11111 --> 0b111110)
        # and invert the encoding if the original number was negative
        # the bit added by the shift will be inverted and that allows
        # to know that the number is negative
        difference = ~(difference << 1) if difference < 0 else difference << 1

        # process 5-bit chunks from right to left until we get a value that
        # is smaller than 0b100000/0x20/32
        while difference >= 0x20:
            # use 0x1f (0b11111) as bitmask to get the smallest 5-bit chunk
            # and 0x20 is used as continuation bit to help decode
            compressed_values += (chr((0x20 | (difference & 0x1f)) + 63))
            # right bitwise shift to exclude the 5-bit chunk that was encoded
            difference >>= 5

        # encode the last 5-bit chunk
        compressed_values += (chr(difference + 63))
        # store number to subtract from next
        previous_num = number

    return compressed_values


def decompress_number(text, index):
    """Decode a single number from a string created with polyline encoding.

    Parameters
    ----------
    text : str
        String representing a compressed list of numbers.
    index : int
        Index of the first character to start decoding a number.

    Returns
    -------
    Index to start decoding the next number and the number decoded
    in the current function call.
    """
    number = 0
    bitwise_shift = 0

    while True:
        # subtract 63 and remove bit from OR with 0x20 if 5-bit chunk
        # is not the last to decode a number that was in the original list
        n = (ord(text[index]) - 63)
        index += 1
        # only continue if there is a continuation bit (0b100000)
        # e.g.: 0b11111 only has 5 bits, meaning it is the last chunk
        # that is necessary to decode the number
        if n >= 0x20:
            # subtract 0x20 (0b100000) to get the 5-bit chunk
            n -= 0x20
            # contruct the binary number with biwise shift to add each
            # 5-bit chunk to original position
            number = number | (n << bitwise_shift)
            # increment bitwise shift value for next 5-bit chunk
            bitwise_shift += 5
        else:
            break

    # add the last chunk, without continuation bit, to the leftmost position
    number = number | (n << bitwise_shift)

    # invert bits to get negative number if sign bit is 1
    # remove sign bit and keep only bits for decoded value
    return index, (~number >> 1) if (number & 1) != 0 else (number >> 1)


def polyline_decoding(text):
    """Decode a list of integers compressed with polyline encoding.

    Parameters
    ----------
    text : str
        String representing a compressed list of numbers.

    Returns
    -------
    number_list : list
        List with the decoded numbers.
    """
    number_list = []
    index = last_num = 0
    # decode precision value
    precision = ord(text[index]) - 63
    index += 1
    while index < len(text):
        # decode a number and get index to start decoding next number
        index, difference = decompress_number(text, index)
        # add decoded difference to get next number in original list
        last_num += difference
        number_list.append(last_num)

    number_list = [round(item * (10 ** (-precision)), precision) for item in number_list]

    return number_list


def replace_list_values(input_list, replace_dict):
    """Replace values in list based on provided substitutions.

    Parameters
    ----------
    input_list : list
        List with values to substitute.
    replace_dict : dict
        Mapping between the values to substitute
        and the values to substitute by.

    Returns
    -------
    replaced_list : list
        List with substituted values (values
        that are not in `replace_dict` are kept
        unchanged).
    """
    replaced_list = [replace_dict.get(e, e) for e in input_list]

    return replaced_list


def replace_chars(column, missing_replace='0'):
    """Replace all non-numeric characters in a column with allele identifiers.

    Parameters
    ----------
    column : pandas.core.series.Series
        Pandas dataframe column.

    Returns
    -------
    replace_missing : pandas.core.series.Series
        Input column with cells that only contain
        numeric characters.
    """
    # remove 'INF-' from inferred alleles
    replace_inf = column.replace(to_replace='INF-',
                                 value='', regex=True)
    # replace '*' in novel alleles from schemas in Chewie-NS
    # before replacing missing data cases to avoid replacing '*' with '0'
    replace_inf = replace_inf.replace(to_replace='\*',
                                      value='', regex=True)
    # replace missing data with
    replace_missing = replace_inf.replace(to_replace='\D+.*',
                                          value=missing_replace, regex=True)

    return replace_missing


def inclusive_range(start, stop, step):
    """
    """

    i = start
    while i < stop:
        yield i
        i += step
    yield stop



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This module contains functions/classes related to parameter
and argument validation.

Code documentation
------------------
"""


import os
import re
import sys
import csv
import shutil
import hashlib
import argparse
import platform
import subprocess
import multiprocessing

try:
    from utils import (constants as ct,
                       file_operations as fo,
                       chewiens_requests as cr,
                       fasta_operations as fao)
except ModuleNotFoundError:
    from CHEWBBACA.utils import (constants as ct,
                                 file_operations as fo,
                                 chewiens_requests as cr,
                                 fasta_operations as fao)


class ModifiedHelpFormatter(argparse.ArgumentDefaultsHelpFormatter):

    # prog is the name of the program 'ex: chewBBACA.py'
    def __init__(self, prog, indent_increment=2, max_help_position=56, width=100):
        super().__init__(prog, indent_increment, max_help_position, width)

    # override split lines method
    def _split_lines(self, text, width):
        lines = super()._split_lines(text, width) + ['']
        return lines

    def _format_action_invocation(self, action):
        if not action.option_strings:
            default = self._get_default_metavar_for_positional(action)
            metavar, = self._metavar_formatter(action, default)(1)
            return metavar

        else:
            parts = []

            # if the Optional doesn't take a value, format is:
            #    -s, --long
            if action.nargs == 0:
                parts.extend(action.option_strings)

            # if the Optional takes a value, format is:
            #    -s ARGS, --long ARGS
            else:
                default = self._get_default_metavar_for_optional(action)
                args_string = self._format_args(action, default)
                for option_string in action.option_strings:
                    parts.append(option_string)

                return '%s %s' % (', '.join(parts), args_string)

            return ', '.join(parts)

    def _get_default_metavar_for_optional(self, action):
        return action.dest.upper()


def arg_list(arg, arg_name):
    """Determine if more than one value has been used for a single parameter.

    Parameter
    ---------
    arg : list
        List with the values used for a sinlge parameter to perform
        allele calling.
    arg_name : str
        The name of the parameter to include in the exception
        message if more than one parameter value has been used
        to perform allele calling.

    Returns
    -------
    The single parameter value used to perform allele calling.

    Raises
    ------
    SystemExit
        - If more than one parameter value has been used to
        perform allele calling.
    """
    if isinstance(arg, list) is True:
        if len(arg) > 1:
            sys.exit('\nMultiple {0} values.'.format(arg_name))
        else:
            arg = arg[0]

    return arg


def bsr_type(arg, min_value=ct.BSR_MIN, max_value=ct.BSR_MAX):
    """Validate the BLAST Score Ratio (BSR) value passed to chewBBACA.

    Parameters
    ----------
    arg : float
        The BLAST Score Ratio (BSR) value passed to chewBBACA.
    min_value : float
        Minimum acceptable BSR value.
    max_value : float
        Maximum acceptable BSR value.

    Returns
    -------
    valid : float
        The BSR value passed to chewBBACA, if it is valid.

    Raises
    ------
    SystemExit
        - If the BSR value cannot be converted to float type
        or if it is not contained in the acceptable interval.
    """
    arg = arg_list(arg, 'BLAST Score Ratio')

    try:
        schema_bsr = float(arg)
        if schema_bsr >= min_value and schema_bsr <= max_value:
            valid = schema_bsr
        elif schema_bsr < min_value or schema_bsr > max_value:
            sys.exit('\nBSR value is not contained in the '
                     '[0.0, 1.0] interval.')
    except Exception:
        sys.exit('\nInvalid BSR value of {0}. BSR value must be contained'
                 ' in the [0.0, 1.0] interval.'.format(arg))

    return valid


def minimum_sequence_length_type(arg, min_value=ct.MSL_MIN, max_value=ct.MSL_MAX):
    """Validate the minimum sequence length value (MSL) passed to chewBBACA.

    Parameters
    ----------
    arg : int
        The MSL value passed to chewBBACA.
    min_value : int
        Minimum acceptable MSL value.
    max_value : int
        Maximum acceptable MSL value.

    Returns
    -------
    valid : int
        The MSL value passed to chewBBACA, if it is valid.

    Raises
    ------
    SystemExit
        - If the MSL value cannot be converted to int type
        or if it is not contained in the acceptable interval.
    """
    arg = arg_list(arg, 'minimum sequence length')

    try:
        schema_ml = int(arg)
        if schema_ml >= min_value and schema_ml <= max_value:
            valid = schema_ml
        elif schema_ml < min_value or schema_ml > max_value:
            sys.exit('\nInvalid minimum sequence length value. '
                     'Must be equal or greater than 0.')
    except Exception:
        sys.exit('\nInvalid minimum sequence length value. '
                 'Value must be a positive integer.')

    return valid


def size_threshold_type(arg, min_value=ct.ST_MIN, max_value=ct.ST_MAX):
    """Validate the size threshold value (ST) passed to chewBBACA.

    Parameters
    ----------
    arg : float
        The ST value passed to chewBBACA. Must be of type float
        or NoneType if no size threshold filter should be applied.
    min_value : float
        Minimum acceptable ST value.
    max_value : float
        Maximum acceptable ST value.

    Returns
    -------
    valid : float
        The ST value passed to chewBBACA, if it is valid.

    Raises
    ------
    SystemExit
        - If the ST value cannot be converted to float type
        or if it is not contained in the acceptable interval.
    """
    arg = arg_list(arg, 'size threshold')

    try:
        schema_st = float(arg)
        if schema_st >= min_value and schema_st <= max_value:
            valid = schema_st
        elif schema_st < min_value or schema_st > max_value:
            sys.exit('\nInvalid size threshold value. '
                     'Must be contained in the [0.0, 1.0] interval.')
    except Exception:
        if arg in [None, 'None']:
            valid = None
        else:
            sys.exit('\nInvalid size threshold value used to '
                     'create schema. Value must be None or a '
                     'positive float in the [0.0, 1.0] interval.')

    return valid


def translation_table_type(arg, genetic_codes=ct.GENETIC_CODES):
    """Validate the translation table value (TT) passed to chewBBACA.

    Parameters
    ----------
    arg : int
        The TT value passed to chewBBACA. Must be of type int
        and match the identifier of one of the genetic codes.
    genetic_codes : dict
        Dictionary with genetic codes identifiers as keys and
        descriptions as values.

    Returns
    -------
    valid : int
        The TT value passed to chewBBACA, if it is valid.

    Raises
    ------
    SystemExit
        - If the TT value cannot be converted to int type
        or if it does not match any of the acceptable genetic
        codes.
    """
    arg = arg_list(arg, 'translation table')

    try:
        schema_gen_code = int(arg)
        if schema_gen_code in genetic_codes:
            valid = schema_gen_code
        else:
            valid = False
    except Exception:
        valid = False

    if valid is False:
        # format available genetic codes into list
        lines = ['\t{0}: {1}'.format(k, v) for k, v in genetic_codes.items()]
        gc_table = '\n{0}\n'.format('\n'.join(lines))

        sys.exit('\nInvalid genetic code value.\nValue must correspond to '
                 'one of the accepted genetic codes\n\nAccepted genetic '
                 'codes:\n{0}'.format(gc_table))

    return valid


def validate_ws(arg, min_value=ct.WORD_SIZE_MIN, max_value=ct.WORD_SIZE_MAX):
    """Validate the word size value (WS) passed to chewBBACA.

    Parameters
    ----------
    arg : float
        The WS value passed to chewBBACA.
    min_value : float
        Minimum acceptable WS value.
    max_value : float
        Maximum acceptable WS value.

    Returns
    -------
    valid : float
        The WS value passed to chewBBACA, if it is valid.

    Raises
    ------
    SystemExit
        - If the WS value cannot be converted to float type
        or if it is not contained in the acceptable interval.
    """
    arg = arg_list(arg, 'word size')

    try:
        if arg is None:
            valid = 'None'
        else:
            word_size = int(arg)
            if word_size >= min_value and word_size <= max_value:
                valid = word_size
            else:
                sys.exit('\nWord size for the clustering step '
                         'must be equal or greater than {0} and '
                         'equal or smaller than {1}.'.format(min_value, max_value))
    except Exception:
        sys.exit('\nSchema created with invalid clustering word '
                 'size value.')

    return valid


def validate_cs(arg, min_value=ct.CLUSTERING_SIMILARITY_MIN,
                max_value=ct.CLUSTERING_SIMILARITY_MAX):
    """Validate the clustering similarity value (CS) passed to chewBBACA.

    Parameters
    ----------
    arg : float
        The CS value passed to chewBBACA. Must be of type float.
    min_value : float
        Minimum acceptable CS value.
    max_value : float
        Maximum acceptable CS value.

    Returns
    -------
    valid : float
        The CS value passed to chewBBACA, if it is valid.

    Raises
    ------
    SystemExit
        - If the CS value cannot be converted to float type
        or if it is not contained in the acceptable interval.
    """
    arg = arg_list(arg, 'clustering similarity')

    try:
        if arg is None:
            valid = 'None'
        else:
            cluster_sim = float(arg)
            if cluster_sim >= min_value and cluster_sim <= max_value:
                valid = cluster_sim
            else:
                sys.exit('\nClustering similarity threshold value '
                         'must be contained in the [0.0, 1.0] '
                         'interval.')
    except Exception:
        sys.exit('\nSchema created with invalid clustering '
                 'threshold value.')

    return valid


def validate_rf(arg, min_value=ct.REPRESENTATIVE_FILTER_MIN,
                max_value=ct.REPRESENTATIVE_FILTER_MAX):
    """Validate the representative filter value (RF) passed to chewBBACA.

    Parameters
    ----------
    arg : float
        The RF value passed to chewBBACA. Must be of type float.
    min_value : float
        Minimum acceptable RF value.
    max_value : float
        Maximum acceptable RF value.

    Returns
    -------
    valid : float
        The RF value passed to chewBBACA, if it is valid.

    Raises
    ------
    SystemExit
        - If the RF value cannot be converted to float type
        or if it is not contained in the acceptable interval.
    """
    arg = arg_list(arg, 'representative filter')

    try:
        if arg is None:
            valid = 'None'
        else:
            representative_filter = float(arg)
            if representative_filter >= min_value and representative_filter <= max_value:
                valid = representative_filter
            else:
                sys.exit('\nRepresentative filter threshold value '
                         'must be contained in the [0.0, 1.0] '
                         'interval.')
    except Exception:
        sys.exit('\nSchema created with invalid representative filter value.')

    return valid


def validate_if(arg, min_value=ct.INTRA_CLUSTER_MIN,
                max_value=ct.INTRA_CLUSTER_MAX):
    """Validate the intra-cluster filter value (IF) passed to chewBBACA.

    Parameters
    ----------
    arg : float
        The IF value passed to chewBBACA. Must be of type float.
    min_value : float
        Minimum acceptable IF value.
    max_value : float
        Maximum acceptable IF value.

    Returns
    -------
    valid : float
        The IF value passed to chewBBACA, if it is valid.

    Raises
    ------
    SystemExit
        - If the IF value cannot be converted to float type
        or if it is not contained in the acceptable interval.
    """
    arg = arg_list(arg, 'intra-cluster filter')

    try:
        if arg is None:
            valid = 'None'
        else:
            intraCluster_filter = float(arg)
            if intraCluster_filter >= min_value and intraCluster_filter <= max_value:
                valid = intraCluster_filter
            else:
                sys.exit('\nIntra-cluster filter value '
                         'must be contained in the [0.0, 1.0] '
                         'interval.')
    except Exception:
        sys.exit('\nSchema created with invalid intra-cluster filter '
                 'value.')

    return valid


def validate_ns_url(arg):
    """Verify if the Chewie-NS URL passed to chewBBACA is valid.

    Parameters
    ----------
    arg : str
        Identifier of the Chewie-NS instance or the URL
        to a instance of Chewie-NS.

    Returns
    -------
    ns_url : str
        URL to connect to the instance of Chewie-NS.

    Raises
    ------
    SystemExit
        - If it is not possible to connect to the
        chewie-NS instance.
    """
    if arg in ct.HOST_NS:
        ns_url = ct.HOST_NS[arg]
    else:
        ns_url = arg

    # sync schema has None by default to get ns_url in schema URI
    if ns_url is not None:
        # check if server is up
        conn = cr.check_connection(ns_url)
        if conn is False:
            sys.exit('Failed to establish a connection to the Chewie-NS '
                     'instance at {0}.'.format(ns_url))

    return ns_url


def validate_python_version(minimum_version=ct.MIN_PYTHON):
    """Validate Python version used to run chewBBACA.

    Parameters
    ----------
    minimum_version : tuple
        A tuple with the Puthon version as (MAJOR, MINOR, PATCH).
        According to the rules of Semanting Versioning
        (https://semver.org/).

    Returns
    -------
    python_version : str
        Python version in format "MAJOR.MINOR.PATCH".

    Raises
    ------
    SystemExit
        - If the Python version does not meet minimum requirements
        or it was not possible to determine/detect a version.
    """
    python_version = platform.python_version()

    try:
        assert tuple(map(int, python_version.split('.'))) >= minimum_version[0]
    except AssertionError:
        print('Python version found: {} '.format(python_version))
        print('Please use version Python >= {0}'.format(minimum_version[1]))
        sys.exit(0)

    return python_version


def verify_cpu_usage(cpu_to_use):
    """Verify if the cores/threads value does not exceed available resources.

    Parameters
    ----------
    cpu_to_use : int
        Value provided for the number of CPU cores/threads.

    Returns
    -------
    cpu_to_use : int
        Value of CPU cores/threads that will be used after
        determining if the provided value was safe.
    """
    total_cpu = multiprocessing.cpu_count()

    cpu_to_use = int(cpu_to_use)

    # do not allow a value greater than the number of cores
    if cpu_to_use >= total_cpu:
        print('Warning! You have provided a CPU core count value '
              'that is equal to or exceeds the number of CPU '
              'cores in your system!')
        # define a value that is safe according to the number of
        # available cores/threads
        if total_cpu > 2:
            cpu_to_use = total_cpu - 2
        elif total_cpu == 2:
            cpu_to_use = 1
        print('Resetting to: {0}'.format(cpu_to_use))
    elif cpu_to_use == (total_cpu - 1):
        print('Warning! You have provided a CPU core count value '
              'that is close to the maximum core count of your '
              'machine ({0}/{1}). This may affect your system '
              'responsiveness.'.format(cpu_to_use, total_cpu))

    return cpu_to_use


def is_exe(fpath):
    """Determine if path points to a file and if the file is an executable.

    Parameters
    ----------
    fpath : str
        Path to a file.

    Returns
    -------
    True if the file exists and is executable, False otherwise.
    """
    return os.path.isfile(fpath) and os.access(fpath, os.X_OK)


def which(program):
    """Determine if a program is in PATH.

    Parameters
    ----------
    program : str
        The name used to call the program.

    Returns
    -------
    program_path : str
        Program path added to PATH.
    """
    proc = subprocess.Popen(['which', program],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)

    stdout, stderr = proc.communicate()

    program_path = stdout.decode('utf8').strip()

    return program_path


def check_blast(blast_path, major=ct.BLAST_MAJOR, minor=ct.BLAST_MINOR):
    """Determine if BLAST is installed and validates its version.

    Parameters
    ----------
    blast_path : str or Nonetype
        Path to the directory with BLAST executables or
        NoneType if user did not provide a value.
    major : int
        BLAST minimun MAJOR version.
    minor : int
        BLAST minimum MINOR version.

    Returns
    -------
    blast_path : str
        Path to the directory with BLAST executables.

    Raises
    ------
    SystemExit
        - If the user did not provide a value and BLAST is
        not in PATH.
        - If the user provided a value but that path does not
        contain BLAST executables.
        - If it is not possible to determine the BLAST
        version or if it does not match minimum requirements.
    """
    # search for BLAST in PATH
    if not blast_path:
        blastp_path = which(ct.BLASTP_ALIAS)
        if not blastp_path:
            sys.exit('Could not find BLAST executables in PATH.')
        else:
            blast_path = os.path.dirname(blastp_path)
    # validate user-provided path
    else:
        blastp_path = os.path.join(blast_path, ct.BLASTP_ALIAS)
        executable = is_exe(blastp_path)
        if executable is False:
            sys.exit('Provided path does not contain BLAST executables.')

    # check BLAST version
    try:
        proc = subprocess.Popen([blastp_path, '-version'],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
    except:
        sys.exit('Could not determine BLAST version.\n'
                 'Please verify that BLAST is installed and added to PATH.')

    stdout, stderr = proc.communicate()

    version_string = stdout.decode('utf8')
    version_pattern = r'^blastp:\s(?P<MAJOR>\d+).(?P<MINOR>\d+).(?P<REV>\d+).*'
    blast_version_pat = re.compile(version_pattern)

    match = blast_version_pat.search(version_string)
    if match is None:
        sys.exit('Could not determine BLAST version.')

    version = {k: int(v) for k, v in match.groupdict().items()}
    if version['MAJOR'] < major or (version['MAJOR'] >= major and version['MINOR'] < minor):
        sys.exit('Please update BLAST to version >= {0}.{1}.0'.format(major, minor))

    return blast_path


def check_prodigal(prodigal_path):
    """Determine if Prodigal is installed and added to the PATH.

    Parameters
    ----------
    prodigal_path : str
        Name or path used to call Prodigal.

    Returns
    -------
    True if Prodigal is installed.

    Raises
    ------
    SystemExit
        - If it is not possible to run the command to determine
        the Prodigal version.
    """
    # check Prodigal version
    try:
        proc = subprocess.Popen([prodigal_path, '-v'],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
    except:
        sys.exit('Could not determine Prodigal version '
                 '({0} -v).\nPlease verify that Prodigal '
                 'is installed and added to PATH.'.format(prodigal_path))

    # prodigal version is captured in stderr
    stdout, stderr = proc.communicate()

    return True


def hash_ptf(ptf_path):
    """Determine hash value for a Prodigal training file.

    Parameters
    ----------
    ptf_path : str or None
        Path to the Prodigal training file or None if no
        training file will be used.

    Returns
    -------
    ptf_hash : str
        Blake2b hash computed from file content.
    """
    if ptf_path is not None:
        ptf_hash = fo.hash_file(ptf_path, hashlib.blake2b())
    else:
        ptf_hash = None

    return ptf_hash


def prompt_arguments(ptf_path, blast_score_ratio, translation_table,
                     minimum_length, size_threshold, version):
    """ Detects if a valid argument value was passed and asks
        for a value if the value of any argument is of type
        NoneType. Raises SystemExit if a provided value is not
        valid.
    """
    prompt = ('It seems that your schema was created with chewBBACA '
              '2.1.0 or lower.\nIt is highly recommended that you run '
              'the PrepExternalSchema process to guarantee full '
              'compatibility with the new chewBBACA version.\nIf you '
              'wish to continue, the AlleleCall process will convert '
              'the schema to v{0}, but will not determine if schema '
              'structure respects configuration values.\nDo you wish '
              'to proceed?\n'.format(version))
    proceed = fo.input_timeout(prompt, ct.prompt_timeout)
    if proceed.lower() not in ['y', 'yes']:
        sys.exit('Exited.')

    print('\n')
    # determine parameters default values to include in config file
    if ptf_path is None:
        prompt = ('Full path to the Prodigal training file:\n')
        ptf_path = fo.input_timeout(prompt, ct.prompt_timeout)
        if ptf_path in ['', 'None']:
            ptf_path = None
        else:
            if os.path.isfile(ptf_path) is False:
                sys.exit('Provided path is not a valid file.')
    else:
        if os.path.isfile(ptf_path) is False:
            sys.exit('Provided path for Prodigal training file is '
                     'not a valid.')

    if blast_score_ratio is None:
        prompt = ('BLAST score ratio value:\n')
        blast_score_ratio = fo.input_timeout(prompt, ct.prompt_timeout)

    blast_score_ratio = bsr_type(blast_score_ratio)

    if translation_table is None:
        prompt = ('Translation table value:\n')
        translation_table = fo.input_timeout(prompt, ct.prompt_timeout)

    translation_table = translation_table_type(translation_table)

    if minimum_length is None:
        prompt = ('Minimum length value:\n')
        minimum_length = fo.input_timeout(prompt, ct.prompt_timeout)

    minimum_length = minimum_sequence_length_type(minimum_length)

    if size_threshold is None:
        prompt = ('Size threshold value:\n')
        size_threshold = fo.input_timeout(prompt, ct.prompt_timeout)

    size_threshold = size_threshold_type(size_threshold)

    return [ptf_path, blast_score_ratio, translation_table,
            minimum_length, size_threshold]


def auto_arguments(ptf_path, blast_score_ratio, translation_table,
                   minimum_length, size_threshold):
    """ Detects if a valid argument value was passed and
        selects default config values if the provided value
        is of type NoneType.
    """

    if ptf_path is not None:
        if os.path.isfile(ptf_path) is False:
            sys.exit('Provided path to Prodigal training file '
                     'is not valid.')
    else:
        ptf_path = None

    blast_score_ratio = (blast_score_ratio
                         if blast_score_ratio is not None
                         else ct.DEFAULT_BSR)
    blast_score_ratio = bsr_type(blast_score_ratio)

    translation_table = (translation_table
                         if translation_table is not None
                         else ct.GENETIC_CODES_DEFAULT)
    translation_table = translation_table_type(translation_table)

    minimum_length = (minimum_length
                      if minimum_length is not None
                      else ct.MSL_MIN)
    minimum_length = minimum_sequence_length_type(minimum_length)

    size_threshold = (size_threshold
                      if size_threshold is not None
                      else ct.SIZE_THRESHOLD_DEFAULT)
    size_threshold = size_threshold_type(size_threshold)

    return [ptf_path, blast_score_ratio, translation_table,
            minimum_length, size_threshold]


def upgrade_legacy_schema(ptf_path, schema_directory, blast_score_ratio,
                          translation_table, minimum_length, version,
                          size_threshold, force_continue):
    """Upgrade a legacy schema to current version.

    Determines configuration values and adds them to the
    configuration file. Creates the file with the list of loci in
    the schema.

    Parameters
    ----------
    ptf_path : str or None
        Path to the Prodigal training file or NoneType
        if no value was provided.
    schema_directory : str
        Path to the schema's directory.
    blast_score_ratio : float
        BLAST Score Ratio value.
    translation_table : int
        Translation table integer identifier.
    minimum_length : int
        Minimum sequence length value.
    version : str
        chewBBACA's version.
    size_threshold : float
        Allele size variation threshold.
    force_continue : bool
        True if argument values should be automatically
        validated and set to default values if their type
        is NoneType. False to validate provided values and
        ask for a value if the user did not provide one.

    Returns
    -------
    Valid/selected values for ptf_path, blast_score_ratio,
    translation_table, minimum_length and size_threshold.
    """
    if force_continue is False:
        values = prompt_arguments(ptf_path, blast_score_ratio,
                                  translation_table, minimum_length,
                                  size_threshold, version)
    # forced method, create with default args defined in constants
    else:
        values = auto_arguments(ptf_path, blast_score_ratio,
                                translation_table, minimum_length,
                                size_threshold)

    ptf_path, blast_score_ratio,\
    translation_table, minimum_length, size_threshold = values

    # copy training file to schema directory
    if ptf_path is not None:
        print('\nAdding Prodigal training file to schema...')
        shutil.copy(ptf_path, schema_directory)
        print('Created {0}'.format(os.path.join(schema_directory,
                                                os.path.basename(ptf_path))))

    # determine PTF hash
    ptf_hash = hash_ptf(ptf_path)

    print('\nCreating file with schema configs...')
    # write schema config file
    schema_config = write_schema_config(blast_score_ratio, ptf_hash,
                                        translation_table, minimum_length,
                                        version, size_threshold,
                                        ct.WORD_SIZE_DEFAULT,
                                        ct.WINDOW_SIZE_DEFAULT,
                                        ct.CLUSTERING_SIMILARITY_DEFAULT,
                                        ct.REPRESENTATIVE_FILTER_DEFAULT,
                                        ct.INTRA_CLUSTER_DEFAULT,
                                        schema_directory)
    print('Created {0}'.format(os.path.join(schema_directory,
                                            '.schema_config')))

    print('\nCreating file with list of genes...')
    # create hidden file with genes/loci list
    genes_list_file = write_gene_list(schema_directory)
    print('Created {0}\n'.format(os.path.join(schema_directory,
                                              '.genes_list')))

    return [ptf_path, blast_score_ratio, translation_table,
            minimum_length, size_threshold]


def validate_ptf_path(ptf_path, schema_directory):
    """ Determines if the path to the Prodigal training file
        is valid. Gets the training file in the schema's
        directory if the input path is of type NoneType.

    Parameters
    ----------
    ptf_path : str or NoneType
        Path to the Prodigal training file or NoneType
        if no value was provided.
    schema_directory : str
        Path to the schema's directory.

    Returns
    -------
    ptf_path : str or bool
        Path to the Prodigal training file or False if
        no training file should be used.

    Raises
    ------
    SystemExit
        - If there is more than one training file in
        the schema's directory.
        - If a path was provided and it is not valid.
    """

    if ptf_path is None:
        # deal with multiple training files
        schema_ptfs = [file
                       for file in os.listdir(schema_directory)
                       if file.endswith('.trn')]
        if len(schema_ptfs) > 1:
            sys.exit('Found more than one Prodigal training '
                     'file in schema directory.\nPlease maintain '
                     'only the training file used in the schema '
                     'creation process.')
        elif len(schema_ptfs) == 1:
            if schema_ptfs[0] is not None:
                ptf_path = os.path.join(schema_directory, schema_ptfs[0])
            else:
                print('There is no Prodigal training file in schema\'s '
                      'directory.')
                ptf_path = None
    else:
        if os.path.isfile(ptf_path) is False:
            message = ('Cannot find specified Prodigal training file.'
                       '\nPlease provide a valid training file.\n\nYou '
                       'can create a training file for a species of '
                       'interest with the following command:\n  prodigal '
                       '-i <reference_genome> -t <training_file.trn> -p '
                       'single\n\nIt is strongly advised to provide a '
                       'high-quality and closed genome for the training '
                       'process.')
            sys.exit(message)

    return ptf_path


def validate_ptf_hash(ptf_hash, schema_ptfs, force_continue):
    """ Determines if the hash for the Prodigal training
        file matches any of the hashes from training files
        that have been used with the schema.

    Paramters
    ---------
    ptf_hash : str
        BLAKE2b hash computed based on the contents of
        the training file.
    schema_ptfs : list
        List with the hashes of all training files that
        have been used with the schema.
    force_continue : bool
        True if the hash should be added to the list with
        all hashes from training files used with the
        schema without prompting the user. False otherwise.

    Returns
    -------
    unmatch : bool
        True if the hash is not in the list with all hashes
        from all training files used with the schema. False
        otherwise.

    Raises
    ------
    SystemExit
        - If the user does not agree to add the hash from a
        new training file to the list with all hashes for
        training files that have been used with the schema.
    """

    unmatch = False
    if ptf_hash not in schema_ptfs:
        ptf_num = len(schema_ptfs)
        if force_continue is False:
            if ptf_num == 1:
                print('Prodigal training file is not the one '
                      'used to create the schema.')
                prompt = ('Using this training file might lead to '
                          'results not consistent with previous runs '
                          'and invalidate the schema for usage with '
                          'the NS.\nContinue process?\n')
                ptf_answer = fo.input_timeout(prompt, ct.prompt_timeout)
            if ptf_num > 1:
                print('Prodigal training file is not any of the {0} '
                      'used in previous runs.'.format(ptf_num))
                prompt = ('Continue?\n')
                ptf_answer = fo.input_timeout(prompt, ct.prompt_timeout)
        else:
            ptf_answer = 'yes'

        if ptf_answer.lower() not in ['y', 'yes']:
            sys.exit('Exited.')
        else:
            unmatch = True

    return unmatch


def validate_ptf(ptf_path, schema_directory, schema_ptfs, force_continue):
    """ Validates the path to the Prodigal training file and
        its hash value.

    Parameters
    ----------
    ptf_path : str or NoneType
        Path to the Prodigal training file or NoneType
        if no value was provided.
    schema_directory : str
        Path to the schema's directory.
    schema_ptfs : list
        List with the hashes of all training files that
        have been used with the schema.
    force_continue : bool
        True if the path and hash of the training file
        should be validated without prompting the user.
        False otherwise.

    Returns
    -------
    ptf_path : str or bool
        Path to the training file if the user provided a
        valid path or if no value was provided and the
        schema has a training file. False if the user
        passed 'False' or if no value was passed and the
        schema has no training file.
    ptf_hash : str
        BLAKE2b hash computed based on the contents of
        the training file.
    unmatch : bool
        True if the training file does not match any of
        the training files previously used with the schema.
    """

    ptf_path = validate_ptf_path(ptf_path, schema_directory)

    # determine PTF checksum
    if ptf_path is not None:
        ptf_hash = hash_ptf(ptf_path)
    else:
        ptf_hash = None

    unmatch = validate_ptf_hash(ptf_hash, schema_ptfs, force_continue)

    return [ptf_path, ptf_hash, unmatch]


def solve_conflicting_arguments(schema_params, ptf_path, blast_score_ratio,
                                translation_table, minimum_length,
                                size_threshold, force_continue, config_file,
                                schema_directory):
    """ Compares schema parameters values stored in the config
        file with values provided by the user to solve conflicting
        cases. Adds/appends new values to the config file if the
        user wants to use values that do not match schema's
        default values.

    Parameters
    ----------
    schema_params : dict
        Dictionary with the schema's config values.
    ptf_path : str or NoneType
        Path to the Prodigal training file or NoneType
        if no value was passed through the command line.
    blast_score_ratio : float or NoneType
        BLAST Score Ratio value. NoneType if no value was
        passed.
    translation_table : int
        Translation table value. NoneType if no value was
        passed.
    minimum_length : int
        Minimum sequence length value. NoneType if no value was
        passed.
    size_threshold : float
        Allele size variation threshold. NoneType if no value was
        passed.
    force_continue : bool
        True to validate parameters values without prompting users.
        False otherwise.
    config_file : str
        Path to the schema's configuration file.
    schema_directory : str
        Path to the schema's directory.

    Returns
    -------
    run_params : dict
        Dictionary with the arguments validated values that
        will be used for allele calling.
    """

    # run parameters values
    run_params = {'bsr': blast_score_ratio,
                  'translation_table': translation_table,
                  'minimum_locus_length': minimum_length,
                  'size_threshold': size_threshold}

    # determine user provided arguments values that differ from default
    unmatch_params = {k: v
                      for k, v in run_params.items()
                      if v not in schema_params[k] and v is not None}
    # determine arguments values not provided by user
    default_params = {k: schema_params[k][0]
                      for k, v in run_params.items()
                      if v is None}

    # update arguments for current run
    for k in run_params:
        if k in default_params:
            run_params[k] = default_params[k]

    if len(unmatch_params) > 0:
        print('Provided arguments values differ from arguments '
              'values used for schema creation:\n')
        params_diffs = [[p, ':'.join(map(str, schema_params[p])),
                         str(unmatch_params[p])]
                        for p in unmatch_params]
        params_diffs_text = ['{:^20} {:^20} {:^10}'.format('Argument', 'Schema', 'Provided')]
        params_diffs_text += ['{:^20} {:^20} {:^10}'.format(p[0], p[1], p[2]) for p in params_diffs]
        print('\n'.join(params_diffs_text))
        if force_continue is False:
            prompt = ('\nContinuing might lead to results not '
                      'consistent with previous runs.\nProviding '
                      'parameters values that differ from the ones '
                      'used for schema creation will also invalidate '
                      'the schema for uploading and synchronization '
                      'with Chewie-NS.\nContinue? (yes/no)\n')
            params_answer = fo.input_timeout(prompt, ct.prompt_timeout)
        else:
            params_answer = 'yes'

        if params_answer.lower() not in ['y', 'yes']:
            sys.exit('Exited.')
        else:
            # append new arguments values to configs values
            for p in unmatch_params:
                schema_params[p].append(unmatch_params[p])

    # default is to get the training file in schema directory
    schema_ptfs = schema_params['prodigal_training_file']
    ptf_path, ptf_hash, unmatch = validate_ptf(ptf_path, schema_directory,
                                               schema_ptfs, force_continue)

    run_params['ptf_path'] = ptf_path
    if unmatch is True:
        schema_params['prodigal_training_file'].append(ptf_hash)
        unmatch_params['prodigal_training_file'] = ptf_hash

    # save updated schema config file
    if len(unmatch_params) > 0:
        fo.pickle_dumper(schema_params, config_file)

    return run_params


def write_gene_list(schema_dir):
    """ Creates list with gene files in a schema and
        uses the pickle module to save the list to a file.

    Parameters
    ----------
    schema_dir : str
        Path to the directory with schema files.

    Returns
    -------
    A list with two elements. A boolean value that
    is True if the file with the list of genes was
    created and False otherwise. The second element
    is the path to the created file.
    """

    schema_files = [file for file in os.listdir(schema_dir) if '.fasta' in file]
    schema_list_file = fo.join_paths(schema_dir, ['.genes_list'])
    fo.pickle_dumper(schema_files, schema_list_file)

    return [os.path.isfile(schema_list_file), schema_list_file]


def write_schema_config(blast_score_ratio, ptf_hash, translation_table,
                        minimum_sequence_length, chewie_version, size_threshold,
                        word_size, window_size, clustering_sim, representative_filter,
                        intra_filter, output_directory):
    """ Writes chewBBACA's parameters values used to create
        a schema to a file.

    Parameters
    ----------
    blast_score_ratio : float
        BLAST Score Ratio value used to create the
        schema.
    ptf_hash : str
        BLAKE2 hash of the Prodigal training file
        content.
    translation_table : int
        Genetic code used to predict and translate
        coding sequences.
    minimum_sequence_length : int
        Minimum sequence length, sequences with a
        length value lower than this value are not
        included in the schema.
    chewie_version : str
        Version of the chewBBACA suite used to create
        the schema.
    size_threshold : float
        Sequence size variation percentage threshold,
        new alleles cannot have a length value that
        deviates +/- than this value in relation to the
        locus's representative sequence.
    word_size : int
        Word/k value used to cluster protein sequences
        during schema creation and allele calling.
    clustering_sim : float
        Proportion of k-mers/minimizers that two proteins
        need to have in common to be clustered together.
    representative_filter : float
        Proportion of k-mers/minimizers that a clustered
        protein has to have in common with the representative
        protein of the cluster to be considered the same gene.
    intra_filter : float
        Proportion of k-mers/minimizers that clustered
        proteins have to have in common to be considered
        of the same gene.
    output_directory : str
        Path to the output directory where the file with
        schema parameters values will be created.

    Returns
    -------
    A list with two elements. A boolean value that
    is True if the file with the parameters values was
    created and False otherwise. The second element
    is the path to the created file.
    """

    size_threshold = None if size_threshold in [None, 'None'] else float(size_threshold)

    params = {}
    params['bsr'] = [float(blast_score_ratio)]
    params['prodigal_training_file'] = [ptf_hash]
    params['translation_table'] = [int(translation_table)]
    params['minimum_locus_length'] = [int(minimum_sequence_length)]
    params['chewBBACA_version'] = [chewie_version]
    params['size_threshold'] = [size_threshold]
    params['word_size'] = [word_size]
    params['window_size'] = [window_size]
    params['cluster_sim'] = [clustering_sim]
    params['representative_filter'] = [representative_filter]
    params['intraCluster_filter'] = [intra_filter]

    config_file = os.path.join(output_directory, '.schema_config')
    fo.pickle_dumper(params, config_file)

    return [os.path.isfile(config_file), config_file]


def read_configs(schema_path, filename):
    """ Reads file with schema config values.

    Parameters
    ----------
    schema_path : str
        Path to the schema's directory.
    filename : str
        Name of the file that contains the config values.

    Returns
    -------
    configs : dict
        Dictionary with config names as keys and config
        values as values.
    """

    config_file = os.path.join(schema_path, filename)
    if os.path.isfile(config_file):
        # Load configs dictionary
        configs = fo.pickle_loader(config_file)
    else:
        sys.exit('Could not find a valid config file.')

    return configs


def check_input_type(input_path, output_file, parent_dir=None):
    """ Checks if the input path is for a file or for a
        directory. If the path is for a directory, the
        function creates a file with the list of paths
        to FASTA files in the directory.

    Parameters
    ----------
    input_path : str
        Path to file or directory.
    output_file : str
        Path to the output file with the list of FASTA
        files.
    parent_dir : str
        Parent directory to add to construct paths
        to input files when users provide a file
        with file names.

    Returns
    -------
    output_file : str
        Path to a file with a list of paths for FASTA
        files.

    Raises
    ------
    SystemExit
        - If there were no FASTA files in the directory.
        - If input path is not a valid path for a file or
          for a directory.
    """

    # check if input argument is a file or a directory
    if os.path.isfile(input_path):
        # check if it's not a FASTA file
        if fao.validate_fasta(input_path) is True:
            sys.exit('Input file is a FASTA file. Please provide '
                     'the path to the parent directory that contains '
                     'the FASTA files or a file with the list of full '
                     'paths to the FASTA files (one per line).')

        # read list of input files
        with open(input_path, 'r') as infile:
            lines = list(csv.reader(infile))
            lines = [f[0] for f in lines]

        # list of input genes must have full paths
        if parent_dir is not None:
            # add parent directory path if necessary
            lines = [os.path.join(parent_dir, f)
                     if parent_dir not in f
                     else f
                     for f in lines]
            # add FASTA extension if it's missing
            lines = [file+'.fasta'
                     if file.endswith('.fasta') is False
                     else file
                     for file in lines]

        # check that all files exist
        missing = [file for file in lines
                   if os.path.exists(file) is False]
        if len(missing) > 0:
            sys.exit('Could not find some of the files provided in '
                     'the input list. Please verify that you\'ve '
                     'provided the full paths to valid input '
                     'files.\n{0}'.format('\n'.join(missing)))
        # save file paths to output file
        else:
            with open(output_file, 'w') as outfile:
                outfile.write('\n'.join(lines))

    elif os.path.isdir(input_path):
        # we need to get only files with FASTA extension
        files = os.listdir(input_path)
        files = fo.filter_files(files, ct.FASTA_SUFFIXES)
        # get absolute paths
        files = [os.path.join(input_path, file) for file in files]
        # filter any directories that might end with FASTA extension
        files = [file for file in files if os.path.isdir(file) is False]

        # only keep files whose content is typical of a FASTA file
        fasta_files = fao.filter_non_fasta(files)

        # if there are FASTA files
        if len(fasta_files) > 0:
            # store full paths to FASTA files
            with open(output_file, 'w') as f:
                for file in fasta_files:
                    f.write(file + '\n')
        else:
            sys.exit('\nCould not get input files. Please '
                     'provide a directory with FASTA files '
                     'or a file with the list of full paths '
                     'to the FASTA files and ensure that '
                     'filenames end with one of the '
                     'following suffixes: {0}.'
                     ''.format(ct.FASTA_SUFFIXES))
    else:
        sys.exit('\nInput argument is not a valid directory or '
                 'file with a list of paths. Please provide a '
                 'valid input, either a folder with FASTA files '
                 'or a file with the list of full paths to FASTA '
                 'files (one per line).')

    return output_file



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This module contains functions related with gene prediction
with Prodigal and extraction of predicted coding sequences
from FASTA files.

Code documentation
------------------
"""


import os
import subprocess

try:
    from utils import (constants as ct,
                       file_operations as fo,
                       fasta_operations as fao,
                       iterables_manipulation as im)
except ModuleNotFoundError:
    from CHEWBBACA.utils import (constants as ct,
                                 file_operations as fo,
                                 fasta_operations as fao,
                                 iterables_manipulation as im)


def extract_genome_cds(reading_frames, contigs, starting_id):
    """Extract coding sequence from contigs.

    Extracts coding sequences from FASTA files based on the
    start and stop coordinates predicted by Prodigal.

    Parameters
    ----------
    reading_frames : str
        Path to the pickled file with the coordinates to
        extract the CDSs predicted by Prodigal.
    contigs : dict
        Dictionary with contig identifiers as keys and
        contig sequences as values.
    starting_id : int
        Integer identifier attributed to the first CDS
        and that will be incremented to serve as identifier
        for subsequent CDSs.

    Returns
    -------
    coding_sequences : dict
        Dictionary with coding sequences ids as keys and
        coding sequences as values.
    coding_sequences_info : list
        List with a sublist for each extracted CDS. Sublists
        have information about the extracted CDS (identifier
        of the contig the CDS was identified in, start position
        in the contig, stop position in the contig, sequence
        identifier attributed to that CDS and the strand that
        coded for that CDS).
    """
    seqid = starting_id
    coding_sequences = {}
    coding_sequences_info = []
    for contig_id, frames in reading_frames.items():
        sequence = contigs[contig_id]
        # for each start and stop codon in the contig
        for cds in frames:
            # start position is 0-based, stop position is upper-bound exclusive
            start_pos = cds[0]
            stop_pos = cds[1]
            strand = cds[2]
            # extract CDS sequence
            cds_sequence = im.extract_single_cds(sequence, *cds).upper()
            seq_hash = im.hash_sequence(cds_sequence)

            # store CDS with unique id
            coding_sequences[seqid] = cds_sequence

            # store CDS information
            coding_sequences_info.append([contig_id, str(start_pos),
                                          str(stop_pos), str(seqid),
                                          str(strand), seq_hash])

            # increment seqid
            seqid += 1

    return [coding_sequences, coding_sequences_info]


def write_coordinates_tsv(cds_info, genome_id, output_file):
    """Write a TSV file with coding sequence coordinates.

    Parameters
    ----------
    cds_info : list
        List with information about each coding sequence
        identified in the genome (contig identifier,
        CDS start position, CDS stop position, CDS
        identifier and CDS coding strand).
    genome_id : str
        Identifier of the genome to add to first field
        of every new line.
    output_file : str
        Path to the output file to which info will
        be saved.
    """
    # write TSV file
    table_lines = [[genome_id] + protein_info[:-1]
                   for protein_info in cds_info]
    table_lines = [im.join_list(line, '\t') for line in table_lines]
    table_text = im.join_list(table_lines, '\n')
    fo.write_to_file(table_text, output_file, 'a', '\n')


def write_coordinates_pickle(cds_info, contig_lengths, output_file):
    """Save coordinates for coding sequences predicted by Prodigal.

    Parameters
    ----------
    cds_info : list
        List with information about each coding sequence
        identified in the genome (contig identifier,
        CDS start position, CDS stop position, CDS
        identifier and CDS coding strand).
    contig_lengths : dict
        Dictionary with contig identifiers as keys and
        contig lengths as values.
    output_file : str
        Path to the output file to which info will
        be saved.
    """
    # write pickle with CDS hash to CDS info dictionary
    pickle_data = [{}, contig_lengths]
    # create dictionary to map CDS hash to CDS location
    for p in cds_info:
        # make sure to store CDS duplicated in the genome
        pickle_data[0].setdefault(p[-1], []).append(p[:-1])

    fo.pickle_dumper(pickle_data, output_file)


def save_extracted_cds(genome, identifier, orf_file, protein_table, cds_file):
    """Extract coding sequences from a FASTA file.

    Extracts coding sequences from a FASTA file based on
    Prodigal's predictions. Writes coding sequences to a
    FASTA file and information about coding sequences to
    a TSV file.

    Parameters
    ----------
    genome : str
        Path to the FASTA file with the FASTA sequences for
        a genome.
    identifier : str
        Genome identifier to add to FASTA records headers
        and to the first field in the TSV file.
    orf_file : str
        Path to the file with Prodigal results.
    protein_table : str
        Path to the TSV file to which coding sequences
        information will be written.
    cds_file : str
        Path to the FASTA file to which coding sequences
        will be written.

    Returns
    -------
    total_cds : int
        Total number of coding sequences extracted from
        the genome.
    """
    # import contigs for current genome/assembly
    contigs = fao.import_sequences(genome)
    # determine contig lengths
    contig_lengths = {k: len(v) for k, v in contigs.items()}
    # extract coding sequences from contigs
    reading_frames = fo.pickle_loader(orf_file)
    genome_info = extract_genome_cds(reading_frames, contigs, 1)
    # save coding sequences to file
    # create records and write them to file
    cds_data = [[identifier, k, v] for k, v in genome_info[0].items()]
    cds_lines = fao.fasta_lines(ct.FASTA_CDS_TEMPLATE, cds_data)
    fo.write_lines(cds_lines, cds_file, write_mode='a')

    write_coordinates_tsv(genome_info[1], identifier, protein_table)
    pickle_out = os.path.join(os.path.dirname(protein_table), identifier+'.cds_hash')
    write_coordinates_pickle(genome_info[1], contig_lengths, pickle_out)

    total_cds = len(genome_info[0])

    return total_cds


def cds_batch_extractor(genomes, index, prodigal_path, temp_directory):
    """Extract coding sequences from a set of genomes.

    Parameters
    ----------
    input_data : list
        List with a set of paths for FASTA files with
        genomic sequences, followed by the path to the
        directory with files with Prodigal resutls, the
        path to the temporary directory for all files and
        directories that will be read and written and
        an index/identifier to add to the output files
        with coding sequences and coding sequences info.

    Returns
    -------
    A list with the following elements:
        protein_table : str
            Path to the TSV file to which coding sequence
            coordinates was written.
        cds_file : str
            Path to the FASTA file to which coding sequences
            were written.
        batch_total : int
            Total number of coding sequences extracted from
            the set of input genomes.
    """
    protein_table = fo.join_paths(temp_directory,
                                  ['cds_coordinates_{0}.tsv'.format(index)])

    cds_file = fo.join_paths(temp_directory,
                             ['coding_sequences_{0}.fasta'.format(index)])

    batch_total = 0
    for g in genomes:
        # determine Prodigal ORF file path for current genome
        identifier = fo.file_basename(g, False)
        orf_file_path = fo.join_paths(prodigal_path,
                                      ['{0}.cds_coordinates'.format(identifier)])
        total = save_extracted_cds(g, identifier, orf_file_path,
                                   protein_table, cds_file)
        batch_total += total

    return [protein_table, cds_file, batch_total]


def run_prodigal(input_file, translation_table, mode, ptf_path):
    """Execute Prodigal.

    Parameters
    ----------
    input_file : str
        Path to a FASTA file.
    translation_table : int
        Genetic code.
    mode : str
        Prodigal execution mode ('single' is the default,
        'meta' should be used to predict genes from smaller
        contigs).
    ptf_path : str or None
        Path to the training file.

    Returns
    -------
    stdout : bytes
        Prodigal's stdout.
    stderr : bytes
        Prodigal's stderr.
    """
    if ptf_path is not None:
        proc = subprocess.Popen(['prodigal', '-i', input_file, '-c',
                                 '-m', '-g', str(translation_table), '-p',
                                 mode, '-f', 'sco', '-q', '-t', ptf_path],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
    elif ptf_path is None:
        proc = subprocess.Popen(['prodigal', '-i', input_file, '-c',
                                 '-m', '-g', str(translation_table), '-p',
                                 mode, '-f', 'sco', '-q'],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)

    # Read the stdout from Prodigal
    stdout = proc.stdout.readlines()
    stderr = proc.stderr.readlines()

    return [stdout, stderr]


def main(input_file, output_dir, ptf_path, translation_table, mode):

    stdout, stderr = run_prodigal(input_file, translation_table, mode, ptf_path)

    # this has to be changed to include the full file name!
    genome_basename = fo.file_basename(input_file, False)

    if len(stderr) > 0:
        stderr = [line.decode('utf-8').strip() for line in stderr]
        stderr = [line for line in stderr if line != '']
        error = ' '.join(stderr)
        return [input_file, error]

    # Parse output
    lines = [line.decode('utf-8').strip() for line in stdout]

    # determine contigs headers indexes
    contigs_headers = [line for line in lines if 'seqhdr' in line]
    contigs_ids = [line.split('"')[1].split()[0] for line in contigs_headers]
    contigs_idx = [lines.index(line) for line in contigs_headers] + [len(lines)]

    # get CDSs' positions for each contig
    contigs_pos = {contigs_ids[i]: lines[contigs_idx[i]+1:contigs_idx[i+1]]
                   for i in range(len(contigs_ids))}

    # exclude contigs without coding sequences
    contigs_pos = {k: v[1:] for k, v in contigs_pos.items() if len(v) > 1}

    # +/1 for sense, -/0 for antisense
    strand_trans = {'+': 1, '-': 0}

    # split and convert list elements
    contigs_pos = {k: [p.split('_')[1:] for p in v]
                   for k, v in contigs_pos.items()}
    contigs_pos = {k: [[int(p[0])-1, int(p[1]), strand_trans[p[2]]]
                   for p in v] for k, v in contigs_pos.items()}

    total_contigs = {k: len(v) for k, v in contigs_pos.items()}
    total_genome = sum(total_contigs.values())

    if total_genome > 0:
        # save positions in file
        filepath = os.path.join(output_dir, genome_basename + '.cds_coordinates')
        fo.pickle_dumper(contigs_pos, filepath)

    status = [input_file, total_genome]

    return status



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This module contains functions related to time measurement
and date formatting.

Code documentation
------------------
"""


import sys
import functools
import datetime as dt


def get_datetime():
    """Return datetime object for current date and hour."""
    current_datetime = dt.datetime.now()

    return current_datetime


def datetime_str(datetime_obj, date_format='%Y-%m-%dT%H:%M:%S'):
    """Convert datetime module object to formatted string.

    Parameters
    ----------
    datetime_obj : datetime.datetime
        Datetime module object.
    date_format : str
        Format for string representation of the date
        object.

    Returns
    -------
    dt_str : str
        String representation of the date according
        to specified format.
    """
    dt_str = dt.datetime.strftime(datetime_obj, date_format)

    return dt_str


def datetime_obj(datetime_str, date_format='%Y-%m-%dT%H:%M:%S'):
    """Convert formatted string to datetime module object.

    Parameters
    ----------
    datetime_str : str
        String to convert to datetime object.
    date_format : str
        Format of the string representation of the date
        object.

    Returns
    -------
    dt_obj : datetime.datetime
        Datetime module object.
    """
    dt_obj = dt.datetime.strptime(datetime_str, date_format)

    return dt_obj


def datetime_diff(sdate, edate):
    """Return the difference in minutes and seconds between two dates.

    Parameters
    ----------
    sdate : datetime.datetime
        Datetime module object corresponding to
        the oldest date.
    edate : datetime.datetime
        Datetime module object corresponding to
        the most recent date.

    Returns
    -------
    A list with the following elements:
        minutes : float
            Time difference between the two dates
            in minutes.
        seconds : float
            The remaining time difference in seconds.
    """
    delta = edate - sdate
    minutes, seconds = divmod(delta.total_seconds(), 60)

    return [minutes, seconds]


def validate_date(date, date_format='%Y-%m-%dT%H:%M:%S.%f'):
    """Check if date is in specified format.

    Parameters
    ----------
    date : str
        String representing a date.
    date_format : str
        Date format.

    Returns
    -------
    valid : str
        The string representing the date if it
        is in the specified format.
    """
    valid = False
    try:
        date = dt.datetime.strptime(date, date_format)
        valid = date
    except ValueError:
        date = dt.datetime.strptime(date+'.0', date_format)
        valid = date

    return valid


def process_header(process):
    """Print a header with the name of the process."""
    header = 'chewBBACA - {0}'.format(process)
    hf = '='*(len(header)+4)
    print('{0}\n  {1}\n{0}'.format(hf, header))


# decorator to time main processes
def process_timer(func):
    # use functools to preserve info about wrapped function
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # get process name and print header
        process_header(sys.argv[1])

        # do not measure time if it is only needed to print the help message
        if any([option in ['-h', '--help'] for option in sys.argv]) is False:
            start = get_datetime()
            start_str = datetime_str(start)
            print('Started at: {0}\n'.format(start_str))

        # run function
        func(*args, **kwargs)

        # does not print elapsed time if the help message is printed
        end = get_datetime()
        end_str = datetime_str(end)
        print('\nFinished at: {0}'.format(end_str))

        minutes, seconds = datetime_diff(start, end)
        print('Took {0: .0f}m{1: .0f}s.'.format(minutes, seconds))

    return wrapper



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Purpose
-------

This module enables the creation of a whole genome multi locus sequence
typing (wgMLST) schema seed.

Expected input
--------------

The process expects the following variables whether through command line
execution or invocation of the :py:func:`main` function:

- ``-i``, ``input_files`` : Path to the directory that contains the input
  FASTA files. Alternatively, a single file with a list of paths to FASTA
  files, one per line.

    - e.g.: ``/home/user/genomes``

- ``-o``, ``output_directory`` : Output directory where the process will
  store intermediate files and create the schema's directory.

    - e.g.: ``/home/user/schemas/new_schema``

- ``--n``, ``schema_name`` : Name given to the folder that will store the
  schema files.

    - e.g.: ``my_schema``

- ``--ptf``, ``ptf_path`` : Path to the Prodigal training file.

    - e.g.: ``/home/user/training_files/species.trn``

- ``--bsr``, ``blast_score_ratio`` : BLAST Score Ratio value.

    - e.g.: ``0.6``

- ``--l``, ``minimum_length`` : Minimum sequence length. Coding sequences
  shorter than this value are excluded.

    - e.g.: ``201``

- ``--t``, ``translation_table`` : Genetic code used to predict genes and
  to translate coding sequences.

    - e.g.: ``11``

- ``--st``, ``size_threshold`` : CDS size variation threshold. Added to the
  schema's config file and used to identify alleles with a length value that
  deviates from the locus length mode during the allele calling process.

    - e.g.: ``0.2``

- ``--w``, ``word_size`` : word size used to generate k-mers during the
  clustering step.

    - e.g.: ``5``

- ``--ws``, ``window_size`` : window size value. Number of consecutive
  k-mers included in each window to determine a minimizer.

    - e.g.: ``5``

- ``--cs``, ``clustering_sim`` : clustering similarity threshold. Minimum
  decimal proportion of shared distinct minimizers for a sequence to be
  added to a cluster.

    - e.g.: ``0.2``

- ``--rf``, ``representative_filter`` : representative similarity threshold.
  Clustered sequences are excluded if they share this proportion of distinct
  minimizers with the cluster representative.

    - e.g.: ``0.9``

- ``--if``, ``intra_filter`` : intra-cluster similarity threshold. Clustered
  sequences are excluded if they share this proportion of distinct minimizers
  with another clustered sequence of equal or greater length.

    - e.g.: ``0.9``

- ``--cpu``, ``cpu_cores`` : Number of CPU cores used to run the process.

    - e.g.: ``4``

- ``--b``, ``blast_path`` : Path to the BLAST executables.

    - e.g.: ``/home/software/blast``

- ``--pm``, ``prodigal_mode`` : Prodigal running mode.

    - e.g.: ``single``

- ``--CDS``, ``cds_input`` : If provided, input is a single or several FASTA
  files with coding sequences (skips gene prediction and CDS extraction).

    - e.g.: ``/home/user/coding_sequences_files``

- ``--no-cleanup``, ``no_cleanup`` : If provided, intermediate files
  generated during process execution are not removed at the end.

Code documentation
------------------
"""


import os
import sys
import math

from Bio import SeqIO

try:
    from utils import (constants as ct,
                       blast_wrapper as bw,
                       core_functions as cf,
                       file_operations as fo,
                       fasta_operations as fao,
                       sequence_manipulation as sm,
                       iterables_manipulation as im,
                       multiprocessing_operations as mo)
except ModuleNotFoundError:
    from CHEWBBACA.utils import (constants as ct,
                                 blast_wrapper as bw,
                                 core_functions as cf,
                                 file_operations as fo,
                                 fasta_operations as fao,
                                 sequence_manipulation as sm,
                                 iterables_manipulation as im,
                                 multiprocessing_operations as mo)


def create_schema_structure(schema_seed_fasta, output_directory, schema_name):
    """Create the schema directory structure.

    Creates the schema seed directory with one FASTA file per
    distinct locus and the `short` directory with the FASTA files
    used to save the representative sequences.

    Parameters
    ----------
    schema_seed_fasta : str
        Path to the FASTA file that contains the sequences that
        constitute the schema seed. Each FASTA record in the file
        is a representative sequence chosen for a locus.
    output_directory : str
        Path to the main output directory of the process.
    schema_name : str
        Name for the schema's directory.

    Returns
    -------
    schema_files : list
        List with the paths to the FASTA files in the schema seed.
    """
    schema_dir = fo.join_paths(output_directory, [schema_name])
    fo.create_directory(schema_dir)

    # add allele identifier to all sequences
    schema_records = {im.replace_multiple_characters(rec.id, ct.CHAR_REPLACEMENTS): str(rec.seq)
                      for rec in SeqIO.parse(schema_seed_fasta, 'fasta')}

    loci_basenames = {k: k+'.fasta' for k in schema_records}
    loci_paths = {k: fo.join_paths(schema_dir, [v])
                  for k, v in loci_basenames.items()}

    for k, v in schema_records.items():
        current_representative = fao.fasta_str_record(ct.FASTA_RECORD_TEMPLATE, [k+'_1', v])
        fo.write_to_file(current_representative, loci_paths[k], 'w', '\n')

    # create 'short' directory
    fo.create_short(loci_paths.values(), schema_dir)

    return loci_paths


def create_schema_seed(fasta_files, output_directory, schema_name, ptf_path,
                       blast_score_ratio, minimum_length, translation_table,
                       size_threshold, word_size, window_size, clustering_sim,
                       representative_filter, intra_filter, cpu_cores, blast_path,
                       prodigal_mode, cds_input):
    """Create a schema seed based on a set of input FASTA files."""
    # map full paths to basename
    inputs_basenames = im.mapping_function(fasta_files,
                                           fo.file_basename, [False])

    # map input identifiers to integers
    # use the mapped integers to refer to each input
    # this reduces memory usage compared to using string identifiers
    basename_map = im.integer_mapping(inputs_basenames.values())
    basename_inverse_map = im.invert_dictionary(basename_map)

    # detect if some inputs share the same unique prefix
    if len(basename_inverse_map) < len(fasta_files):
        basename_counts = [[basename, list(inputs_basenames.values()).count(basename)]
                           for basename in basename_map]
        repeated_basenames = ['{0}: {1}'.format(*l)
                              for l in basename_counts if l[1] > 1]
        fo.delete_directory(output_directory)
        sys.exit('\nSome input files share the same filename prefix '
                 '(substring before the first "." in the filename). '
                 'Please make sure that every input file has a unique '
                 'filename prefix.\n{0}'.format('\n'.join(repeated_basenames)))

    # define directory for temporary files
    temp_directory = fo.join_paths(output_directory, ['temp'])
    fo.create_directory(temp_directory)

    if cds_input is False:
        print('Number of inputs: {0}'.format(len(fasta_files)))

        # create directory to store files with Prodigal results
        prodigal_path = fo.join_paths(temp_directory, ['1_gene_prediction'])
        fo.create_directory(prodigal_path)

        # run Prodigal to determine CDSs for all input genomes
        print('\nPredicting gene sequences...\n')

        # gene prediction step
        failed = cf.predict_genes(fasta_files, ptf_path,
                                  translation_table, prodigal_mode,
                                  cpu_cores, prodigal_path)

        if len(failed) > 0:
            print('\nFailed to predict genes for {0} genomes'
                  '.'.format(len(failed)))
            print('Make sure that Prodigal runs in meta mode (--pm meta) '
                  'if any input file has less than 100kbp.')

            # remove failed genomes from paths
            fasta_files = im.filter_list(fasta_files, failed)

        if len(fasta_files) == 0:
            sys.exit('\nCould not predict gene sequences from any '
                     'of the input files.\nPlease provide input files '
                     'in the accepted FASTA format.')

        # CDS extraction step
        print('\n\nExtracting coding sequences...\n')
        # create output directory
        cds_extraction_path = fo.join_paths(temp_directory,
                                            ['2_cds_extraction'])
        fo.create_directory(cds_extraction_path)
        eg_results = cf.extract_genes(fasta_files, prodigal_path,
                                      cpu_cores, cds_extraction_path)
        cds_files, total_extracted, cds_coordinates = eg_results
        print('\n\nExtracted a total of {0} coding sequences from {1} '
              'genomes.'.format(total_extracted, len(fasta_files)))
        # move TSV file with CDS coordinates to output directory
        fo.move_file(cds_coordinates, output_directory)
    else:
        cds_files = fasta_files
        failed = []
        print('Number of inputs: {0}'.format(len(cds_files)))

    # create directory to store files from pre-process steps
    preprocess_dir = fo.join_paths(temp_directory, ['3_cds_preprocess'])
    fo.create_directory(preprocess_dir)

    # DNA sequences deduplication step
    # keep hash of unique sequences and a list with the integer
    # identifiers of genomes that have those sequences
    # lists of integers are encoded with polyline algorithm
    print('\nRemoving duplicated DNA sequences...', end='')
    # create directory to store files from DNA deduplication
    dna_dedup_dir = fo.join_paths(preprocess_dir, ['cds_deduplication'])
    fo.create_directory(dna_dedup_dir)
    distinct_dna_template = 'distinct_cds_{0}'
    ds_results = cf.exclude_duplicates(cds_files, dna_dedup_dir, cpu_cores,
                                       distinct_dna_template, [basename_map, basename_inverse_map],
                                       False, True)

    distinct_seqids, distinct_file, repeated = ds_results
    print('removed {0} sequences.'.format(int(repeated)))

    print('Kept {0} distinct sequences.'.format(len(distinct_seqids)))

    indexed_dna_file = SeqIO.index(distinct_file, 'fasta')

    # determine small sequences step
    print('\nRemoving sequences smaller than {0} '
          'nucleotides...'.format(minimum_length), end='')
    ss_results = cf.exclude_small(distinct_file, minimum_length)
    small_seqids, ss_lines = ss_results
    print('removed {0} sequences.'.format(len(small_seqids)))

    # exclude seqids of small sequences
    schema_seqids = list(set(distinct_seqids) - set(small_seqids))
    schema_seqids = im.sort_iterable(schema_seqids, sort_key=lambda x: x.lower())

    # sequence translation step
    cds_translation_dir = fo.join_paths(preprocess_dir, ['cds_translation'])
    fo.create_directory(cds_translation_dir)
    print('Translating {0} CDS...'.format(len(schema_seqids)))
    ts_results = cf.translate_sequences(schema_seqids, distinct_file,
                                        cds_translation_dir, translation_table,
                                        minimum_length, cpu_cores)
    protein_file, ut_seqids, ut_lines = ts_results
    print('\nIdentified {0} CDS that could not be translated.'.format(len(ut_seqids)))

    # write info about invalid alleles to file
    invalid_alleles_file = fo.join_paths(output_directory,
                                         ['invalid_cds.txt'])
    invalid_alleles = im.join_list(ut_lines+ss_lines, '\n')
    fo.write_to_file(invalid_alleles, invalid_alleles_file, 'w', '\n')
    print('\nInfo about untranslatable and small sequences '
          'stored in {0}'.format(invalid_alleles_file))

    # protein sequences deduplication step
    # create directory to store files from protein deduplication
    print('\nRemoving duplicated protein sequences...', end='')
    protein_dedup_dir = fo.join_paths(preprocess_dir,
                                      ['translated_cds_deduplication'])
    fo.create_directory(protein_dedup_dir)
    distinct_prot_template = 'distinct_translated_cds_{0}'
    ds_results = cf.exclude_duplicates([protein_file], protein_dedup_dir, 1,
                                       distinct_prot_template,
                                       [basename_map, basename_inverse_map],
                                       True, True)

    distinct_protein_seqs, distinct_prots_file, repeated = ds_results

    schema_seqids = im.sort_iterable(distinct_protein_seqs,
                                     sort_key=lambda x: x.lower())

    print('removed {0} sequences.'.format(int(repeated)))
    print('\nKept {0} sequences after filtering the initial '
          'sequences.'.format(len(distinct_protein_seqs)))

    # protein clustering step
    # read protein sequences
    proteins = fao.import_sequences(distinct_prots_file)

    # create directory to store clustering data
    clustering_dir = fo.join_paths(temp_directory, ['4_clustering'])
    fo.create_directory(clustering_dir)

    print('Clustering proteins...')
    group_size = math.ceil(len(proteins)/40)
    cs_results = cf.cluster_sequences(proteins, word_size, window_size,
                                      clustering_sim, None, True,
                                      1, 1, clustering_dir, cpu_cores,
                                      group_size, False)
    print('\nClustered {0} proteins into {1} clusters.'
          ''.format(len(proteins), len(cs_results)))

    # exclude based on high similarity to cluster representatives
    rep_filter_dir = fo.join_paths(clustering_dir, ['representative_filter'])
    fo.create_directory(rep_filter_dir)
    print('Removing sequences based on high similarity with the '
          'cluster representative...', end='')
    cp_results = cf.cluster_representative_filter(cs_results,
                                                  representative_filter,
                                                  rep_filter_dir)
    clusters, excluded_seqids, singletons, clustered_sequences = cp_results
    print('removed {0} sequences.'.format(len(excluded_seqids)))
    print('Identified {0} singletons.'.format(len(singletons)))
    print('Remaining sequences after representative and singleton '
          'pruning: {0}'.format(clustered_sequences))

    # remove excluded seqids
    schema_seqids = list(set(schema_seqids) - excluded_seqids)

    # exclude based on high similarity to other clustered sequences
    intra_filter_dir = fo.join_paths(clustering_dir, ['intracluster_filter'])
    fo.create_directory(intra_filter_dir)
    print('Removing sequences based on high similarity with '
          'other clustered sequences...', end='')
    cip_results = cf.cluster_intra_filter(clusters, proteins,
                                          word_size, intra_filter,
                                          intra_filter_dir)
    clusters, intra_excluded = cip_results
    print('removed {0} sequences.'.format(len(intra_excluded)))

    # remove excluded seqids - we get set of sequences from clusters
    # plus singletons
    schema_seqids = list(set(schema_seqids) - set(intra_excluded))

    # define BLASTp and makeblastdb paths
    blastp_path = fo.join_paths(blast_path, [ct.BLASTP_ALIAS])
    makeblastdb_path = fo.join_paths(blast_path, [ct.MAKEBLASTDB_ALIAS])

    if len(clusters) > 0:
        blasting_dir = fo.join_paths(clustering_dir, ['cluster_BLASTer'])
        fo.create_directory(blasting_dir)

        print('Clusters to BLAST: {0}'.format(len(clusters)))
        blast_results, ids_dict = cf.blast_clusters(clusters, proteins,
                                                    blasting_dir, blastp_path,
                                                    makeblastdb_path, cpu_cores)

        blast_files = im.flatten_list(blast_results)

        # compute and exclude based on BSR
        blast_excluded_alleles = [sm.apply_bsr(fo.read_tabular(file),
                                               indexed_dna_file,
                                               blast_score_ratio,
                                               ids_dict)
                                  for file in blast_files]

        # merge bsr results
        blast_excluded_alleles = im.flatten_list(blast_excluded_alleles)

        blast_excluded_alleles = set([ids_dict[seqid] for seqid in blast_excluded_alleles])
        schema_seqids = list(set(schema_seqids) - blast_excluded_alleles)
        print('\n\nRemoved {0} sequences based on high BSR value with '
              'other sequences.'.format(len(blast_excluded_alleles)))

        # write list of excluded to file
        blast_excluded_outfile = fo.join_paths(blasting_dir, ['excluded.txt'])
        fo.write_lines(blast_excluded_alleles, blast_excluded_outfile)

    # perform final BLAST to identify similar sequences that do not
    # share many/any kmers
    print('Total of {0} sequences to compare in final BLAST.'.format(len(schema_seqids)))

    # sort seqids before final BLASTp to ensure consistent results
    schema_seqids = im.sort_iterable(schema_seqids, sort_key=lambda x: x.lower())

    # create directory for final BLASTp
    final_blast_dir = fo.join_paths(temp_directory, ['5_final_blast'])
    fo.create_directory(final_blast_dir)

    # create FASTA file with remaining sequences
    beta_file = os.path.join(final_blast_dir, 'remaining_sequences.fasta')
    fao.get_sequences_by_id(proteins, schema_seqids, beta_file)

    # change sequence identifiers to avoid BLAST error
    # related to sequence header ength limite
    integer_seqids = os.path.join(final_blast_dir, 'remaining_sequences_integer_headers.fasta')
    ids_dict2 = fao.integer_headers(beta_file, integer_seqids)

    # create BLASTp database
    blast_db = fo.join_paths(final_blast_dir, ['remaining_sequences'])
    db_stderr = bw.make_blast_db(makeblastdb_path, integer_seqids, blast_db, 'prot')

    if len(db_stderr) > 0:
        sys.exit(db_stderr)

    # divide FASTA file into groups of 100 sequences to reduce
    # execution time for large sequence sets
    split_dir = fo.join_paths(final_blast_dir, ['cds_subsets'])
    fo.create_directory(split_dir)
    splitted_fastas = fao.split_seqcount(integer_seqids, split_dir, 100)

    # create directory to store results from final BLASTp
    final_blastp_dir = fo.join_paths(final_blast_dir, ['BLAST_results'])
    fo.create_directory(final_blastp_dir)
    blast_outputs = ['{0}/{1}_blast_out.tsv'.format(final_blastp_dir,
                                                    fo.file_basename(i[0], False))
                     for i in splitted_fastas]

    # add common arguments to all sublists
    blast_inputs = [[blastp_path, blast_db, file[0],
                     blast_outputs[i], 1, 1, bw.run_blast]
                    for i, file in enumerate(splitted_fastas)]

    print('Performing final BLASTp...')
    blast_stderr = mo.map_async_parallelizer(blast_inputs,
                                             mo.function_helper,
                                             cpu_cores,
                                             show_progress=True)

    blast_stderr = im.flatten_list(blast_stderr)
    if len(blast_stderr) > 0:
        sys.exit(blast_stderr)

    # concatenate files with BLASTp results
    blast_output = fo.join_paths(final_blast_dir, ['blast_out_concat.tsv'])
    blast_output = fo.concatenate_files(blast_outputs, blast_output)

    final_excluded = sm.apply_bsr(fo.read_tabular(blast_output),
                                  indexed_dna_file,
                                  blast_score_ratio,
                                  ids_dict2)
    final_excluded = [ids_dict2[seqid] for seqid in final_excluded]

    schema_seqids = list(set(schema_seqids) - set(final_excluded))

    print('\nRemoved {0} sequences that were highly similar '
          'to other sequences.'.format(len(final_excluded)))

    # create file with the schema representative sequences
    loci_representatives = os.path.join(final_blast_dir, 'loci_representatives.fasta')
    fao.get_sequences_by_id(indexed_dna_file, schema_seqids, loci_representatives)

    schema_files = create_schema_structure(loci_representatives, output_directory,
                                           schema_name)

    return [schema_files, temp_directory, failed]


def main(input_files, output_directory, schema_name, ptf_path,
         blast_score_ratio, minimum_length, translation_table,
         size_threshold, word_size, window_size, clustering_sim,
         representative_filter, intra_filter, cpu_cores, blast_path,
         cds_input, prodigal_mode, no_cleanup):

    print('Prodigal training file: {0}'.format(ptf_path))
    print('CPU cores: {0}'.format(cpu_cores))
    print('BLAST Score Ratio: {0}'.format(blast_score_ratio))
    print('Translation table: {0}'.format(translation_table))
    print('Minimum sequence length: {0}'.format(minimum_length))
    print('Size threshold: {0}'.format(size_threshold))
    print('Word size: {0}'.format(word_size))
    print('Window size: {0}'.format(window_size))
    print('Clustering similarity: {0}'.format(clustering_sim))
    print('Representative filter: {0}'.format(representative_filter))
    print('Intra-cluster filter: {0}'.format(intra_filter))

    # read file with paths to input files
    input_files = fo.read_lines(input_files, strip=True)

    # sort paths to FASTA files
    input_files = im.sort_iterable(input_files, sort_key=lambda x: x.lower())

    results = create_schema_seed(input_files, output_directory, schema_name,
                                 ptf_path, blast_score_ratio, minimum_length,
                                 translation_table, size_threshold, word_size,
                                 window_size, clustering_sim, representative_filter,
                                 intra_filter, cpu_cores, blast_path,
                                 prodigal_mode, cds_input)

    # remove temporary files
    if no_cleanup is False:
        exists = fo.delete_directory(results[1])

    # print message about schema that was created
    print('Created schema seed with {0} loci.'.format(len(results[0])))



from setuptools import setup

import CHEWBBACA

VERSION = CHEWBBACA.__version__

setup(
  name = 'chewBBACA',
  packages = ['CHEWBBACA', 'CHEWBBACA.AlleleCall', 'CHEWBBACA.CreateSchema',
              'CHEWBBACA.utils', 'CHEWBBACA.SchemaEvaluator', 'CHEWBBACA.PrepExternalSchema',
              'CHEWBBACA.CHEWBBACA_NS', 'CHEWBBACA.UniprotFinder', 'CHEWBBACA.ExtractCgMLST'],
  version = VERSION,
  description = 'A complete suite for gene-by-gene schema creation and strain identification.',
  author = 'Mickael Silva, Pedro Cerqueira, Rafael Mamede',
  author_email = 'imm-bioinfo@medicina.ulisboa.pt',
  url = 'https://github.com/B-UMMI/chewBBACA', 
  keywords = ['cgMLST', 'bacterial typing', 'nomenclature server'],
  install_requires = ['numpy>=1.23.4', 'scipy>=1.9.3', 'biopython>=1.78',
                      'plotly>=5.8.0', 'SPARQLWrapper>=2.0.0', 'pandas>=1.5.1',
                      'requests>=2.27.1'],
  python_requires = '>=3.7',
  include_package_data = True,
  entry_points={'console_scripts': ["chewBBACA.py = CHEWBBACA.chewBBACA:main",
                                    "chewie = CHEWBBACA.chewBBACA:main"]
                }
)
